"""
Moving Sprite Stress Test

Simple program to test how fast we can draw sprites that are moving

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.stress_test_draw_moving
"""
import random
import arcade
import timeit
import time
import collections
import pyglet

# --- Constants ---
SPRITE_SCALING_COIN = 0.25
SPRITE_NATIVE_SIZE = 128
SPRITE_SIZE = int(SPRITE_NATIVE_SIZE * SPRITE_SCALING_COIN)
COIN_COUNT_INCREMENT = 1000

STOP_COUNT = 15000
RESULTS_FILE = "stress_test_draw_moving_arcade.csv"

WINDOW_WIDTH = 1800
WINDOW_HEIGHT = 1000
WINDOW_TITLE = "Moving Sprite Stress Test"


class FPSCounter:
    def __init__(self):
        self.time = time.perf_counter()
        self.frame_times = collections.deque(maxlen=60)

    def tick(self):
        t1 = time.perf_counter()
        dt = t1 - self.time
        self.time = t1
        self.frame_times.append(dt)

    def get_fps(self):
        total_time = sum(self.frame_times)
        if total_time == 0:
            return 0
        else:
            return len(self.frame_times) / sum(self.frame_times)


class Coin(arcade.Sprite):

    def update(self, delta_times):
        """
        Update the sprite.
        """
        self.position = (self.position[0] + self.change_x, self.position[1] + self.change_y)


class GameView(arcade.View):
    """ Our custom Window Class"""

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.coin_list = None

        self.processing_time = 0
        self.draw_time = 0
        self.program_start_time = timeit.default_timer()
        self.sprite_count_list = []
        self.fps_list = []
        self.processing_time_list = []
        self.drawing_time_list = []
        self.last_fps_reading = 0
        self.fps = FPSCounter()

        self.background_color = arcade.color.AMAZON

        # Open file to save timings
        self.results_file = open(RESULTS_FILE, "w")

    def add_coins(self):

        # Create the coins
        for i in range(COIN_COUNT_INCREMENT):
            # Create the coin instance
            # Coin image from kenney.nl
            coin = Coin(":resources:images/items/coinGold.png", SPRITE_SCALING_COIN)

            # Position the coin
            coin.center_x = random.randrange(SPRITE_SIZE, WINDOW_WIDTH - SPRITE_SIZE)
            coin.center_y = random.randrange(SPRITE_SIZE, WINDOW_HEIGHT - SPRITE_SIZE)

            coin.change_x = random.randrange(-3, 4)
            coin.change_y = random.randrange(-3, 4)

            # Add the coin to the lists
            self.coin_list.append(coin)

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.coin_list = arcade.SpriteList(use_spatial_hash=False)

    def on_draw(self):
        """ Draw everything """

        # Start timing how long this takes
        draw_start_time = timeit.default_timer()

        self.clear()
        self.coin_list.draw()

        # Display info on sprites
        # output = f"Sprite count: {len(self.coin_list):,}"
        # arcade.draw_text(output, 20, WINDOW_HEIGHT - 20, arcade.color.BLACK, 16)
        #
        # # Display timings
        # output = f"Processing time: {self.processing_time:.3f}"
        # arcade.draw_text(output, 20, WINDOW_HEIGHT - 40, arcade.color.BLACK, 16)
        #
        # output = f"Drawing time: {self.draw_time:.3f}"
        # arcade.draw_text(output, 20, WINDOW_HEIGHT - 60, arcade.color.BLACK, 16)
        #
        # fps = self.fps.get_fps()
        # output = f"FPS: {fps:3.0f}"
        # arcade.draw_text(output, 20, WINDOW_HEIGHT - 80, arcade.color.BLACK, 16)

        self.draw_time = timeit.default_timer() - draw_start_time
        self.fps.tick()

    def on_update(self, delta_time):
        # Start update timer
        start_time = timeit.default_timer()

        self.coin_list.update()

        for sprite in self.coin_list:

            if sprite.position[0] < 0:
                sprite.change_x *= -1
            elif sprite.position[0] > WINDOW_WIDTH:
                sprite.change_x *= -1
            if sprite.position[1] < 0:
                sprite.change_y *= -1
            elif sprite.position[1] > WINDOW_HEIGHT:
                sprite.change_y *= -1

        # Save the time it took to do this.
        self.processing_time = timeit.default_timer() - start_time

        # Total time program has been running
        total_program_time = int(timeit.default_timer() - self.program_start_time)

        # Print out stats, or add more sprites
        if total_program_time > self.last_fps_reading:
            self.last_fps_reading = total_program_time

            # It takes the program a while to "warm up", so the first
            # few seconds our readings will be off. So wait some time
            # before taking readings
            if total_program_time > 5:

                # We want the program to run for a while before taking
                # timing measurements. We don't want the time it takes
                # to add new sprites to be part of that measurement. So
                # make sure we have a clear second of nothing but
                # running the sprites, and not adding the sprites.
                if total_program_time % 2 == 1:

                    # Take timings
                    output = (
                        f"{total_program_time}, "
                        f"{len(self.coin_list)}, "
                        f"{self.fps.get_fps():.1f}, "
                        f"{self.processing_time:.4f}, "
                        f"{self.draw_time:.4f}\n"
                    )

                    self.results_file.write(output)
                    print(output, end="")
                    if len(self.coin_list) >= STOP_COUNT:
                        pyglet.app.exit()
                        return

                    self.sprite_count_list.append(len(self.coin_list))
                    self.fps_list.append(round(self.fps.get_fps(), 1))
                    self.processing_time_list.append(self.processing_time)
                    self.drawing_time_list.append(self.draw_time)

                    # Now add the coins
                    self.add_coins()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



#  type: ignore
from __future__ import annotations

from arcade.future.input import inputs
from arcade.future.input.raw_dicts import RawAction, RawActionMapping, RawAxis, RawAxisMapping


class Action:
    def __init__(self, name: str) -> None:
        self.name = name
        self._mappings: set[ActionMapping] = set()

    def add_mapping(self, mapping: ActionMapping) -> None:
        self._mappings.add(mapping)

    def remove_mapping(self, mapping: ActionMapping) -> None:
        self._mappings.discard(mapping)


class Axis:
    def __init__(self, name: str) -> None:
        self.name = name
        self._mappings: set[AxisMapping] = set()

    def add_mapping(self, mapping: AxisMapping) -> None:
        self._mappings.add(mapping)

    def remove_mapping(self, mapping: AxisMapping) -> None:
        self._mappings.discard(mapping)


class InputMapping:
    """Base class for other mappings.

    Without subclassing, it's still useful as a type annotation since we
    we use type hashing a lot.
    """

    def __init__(self, input: inputs.InputEnum):
        try:
            self._input_type = inputs.CLASS_TO_INPUT_TYPE[type(input)]
        except KeyError:
            raise TypeError(
                f"Got {input} input specified for ActionMapping must be of of: "
                f"{', '.join(t.__name__ for t in inputs.CLASS_TO_INPUT_TYPE.keys())}"
            )
        self._input = input


class ActionMapping(InputMapping):
    def __init__(
        self,
        input: inputs.InputEnum,
        mod_shift: bool = False,
        mod_ctrl: bool = False,
        mod_alt: bool = False,
    ):
        super().__init__(input)
        self._modifiers = set()
        if mod_shift:
            self._modifiers.add(inputs.Keys.MOD_SHIFT.value)
        if mod_ctrl:
            self._modifiers.add(inputs.Keys.MOD_ACCEL.value)
        if mod_alt:
            self._modifiers.add(inputs.Keys.MOD_ALT.value)
            self._modifiers.add(inputs.Keys.MOD_OPTION.value)


class AxisMapping(InputMapping):
    def __init__(self, input: inputs.InputEnum, scale: float):
        super().__init__(input)
        self._scale = scale


def serialize_action(action: Action) -> RawAction:
    raw_mappings: list[RawActionMapping] = []
    for mapping in action._mappings:
        raw_mappings.append(
            {
                "input": mapping._input.value,
                "input_type": mapping._input_type.value,
                "mod_shift": inputs.Keys.MOD_SHIFT in mapping._modifiers,
                "mod_ctrl": inputs.Keys.MOD_ACCEL in mapping._modifiers,
                "mod_alt": (
                    inputs.Keys.MOD_ALT in mapping._modifiers
                    or inputs.Keys.MOD_OPTION in mapping._modifiers
                ),
            }
        )
    return {"name": action.name, "mappings": raw_mappings}


def parse_raw_axis(raw_axis: RawAxis) -> Axis:
    axis = Axis(raw_axis["name"])
    for raw_mapping in raw_axis["mappings"]:
        instance = inputs.parse_mapping_input_enum(raw_mapping)
        axis.add_mapping(
            AxisMapping(
                instance,
                raw_mapping["scale"],
            )
        )

    return axis


def serialize_axis(axis: Axis) -> RawAxis:
    raw_mappings: list[RawAxisMapping] = []
    for mapping in axis._mappings:
        raw_mappings.append(
            {
                "input": mapping._input.value,
                "input_type": mapping._input_type.value,
                "scale": mapping._scale,
            }
        )
    return {"name": axis.name, "mappings": raw_mappings}



"""
Drawing text with pyglet label
"""

from ctypes import c_int, c_ubyte
from pathlib import Path
from typing import Any

import pyglet

import arcade
from arcade.exceptions import PerformanceWarning, warning
from arcade.resources import resolve
from arcade.texture_atlas import TextureAtlasBase
from arcade.types import Color, Point, RGBOrA255

__all__ = ["load_font", "Text", "create_text_sprite", "draw_text"]


class _ArcadeTextLayoutGroup(pyglet.text.layout.TextLayoutGroup):
    """Create a text layout rendering group.

    Overrides pyglet blending handling to allow for additive blending.
    Furthermore, it resets the blend function to the previous state.
    """

    _prev_blend: bool
    _prev_blend_func: tuple[int, int, int, int]

    def set_state(self) -> None:
        self.program.use()
        self.program["scissor"] = False

        pyglet.gl.glActiveTexture(pyglet.gl.GL_TEXTURE0)
        pyglet.gl.glBindTexture(self.texture.target, self.texture.id)

        blend = c_ubyte()
        pyglet.gl.glGetBooleanv(pyglet.gl.GL_BLEND, blend)
        self._prev_blend = bool(blend.value)

        src_rgb = c_int()
        dst_rgb = c_int()
        src_alpha = c_int()
        dst_alpha = c_int()
        pyglet.gl.glGetIntegerv(pyglet.gl.GL_BLEND_SRC_RGB, src_rgb)
        pyglet.gl.glGetIntegerv(pyglet.gl.GL_BLEND_DST_RGB, dst_rgb)
        pyglet.gl.glGetIntegerv(pyglet.gl.GL_BLEND_SRC_ALPHA, src_alpha)
        pyglet.gl.glGetIntegerv(pyglet.gl.GL_BLEND_DST_ALPHA, dst_alpha)

        self._prev_blend_func = (src_rgb.value, dst_rgb.value, src_alpha.value, dst_alpha.value)

        pyglet.gl.glEnable(pyglet.gl.GL_BLEND)
        pyglet.gl.glBlendFuncSeparate(
            pyglet.gl.GL_SRC_ALPHA,
            pyglet.gl.GL_ONE_MINUS_SRC_ALPHA,
            pyglet.gl.GL_ONE,
            pyglet.gl.GL_ONE,
        )

    def unset_state(self) -> None:
        if not self._prev_blend:
            pyglet.gl.glDisable(pyglet.gl.GL_BLEND)

        pyglet.gl.glBlendFuncSeparate(
            self._prev_blend_func[0],
            self._prev_blend_func[1],
            self._prev_blend_func[2],
            self._prev_blend_func[3],
        )
        self.program.stop()


pyglet.text.layout.TextLayout.group_class = _ArcadeTextLayoutGroup


def load_font(path: str | Path) -> None:
    """
    Load fonts in a file (usually .ttf) adding them to a global font registry.

    A file can contain one or multiple fonts. Each font has a name.
    Open the font file to find the actual name(s). These names
    are used to select font when drawing text.

    Examples::

        # Load a font in the current working directory
        # (absolute path is often better)
        arcade.load_font("Custom.ttf")
        # Load a font using a custom resource handle
        arcade.load_font(":font:Custom.ttf")

    Args:
        path: Path to the font file
    Raises:
        FileNotFoundError: if the font specified wasn't found
    """
    file_path = resolve(path)
    pyglet.font.add_file(str(file_path))


FontNameOrNames = str | tuple[str, ...]


def _attempt_font_name_resolution(font_name: FontNameOrNames) -> str:
    """Attempt to resolve a font name.

    Preserves the original logic of this section, even though it
    doesn't seem to make sense entirely. Comments are an attempt
    to make sense of the original code.

    If it can't resolve a definite path, it will return the original
    argument for pyglet to attempt to resolve. This is consistent with
    the original behavior of this code before it was encapsulated.

    Args:
        font_name: A font name, path to a font file, or list of names
    """
    if font_name:
        # ensure
        if isinstance(font_name, str):
            font_list: tuple[str, ...] = (font_name,)
        elif isinstance(font_name, tuple):
            font_list = font_name
        else:
            raise TypeError(
                "font_name parameter must be a string, "
                "or a tuple of strings that specify a font name."
            )

        for font in font_list:
            try:
                path = resolve(font)
                # print(f"Font path: {path=}")

                # found a font successfully!
                return path.name

            except FileNotFoundError:
                pass

        # failed to find it ourselves, hope pyglet can make sense of it
        # Note this is the best approximation of what I understand the old
        # behavior to have been.
        return pyglet.font.load(font_list).name

    raise ValueError(f"Couldn't find a font for {font_name!r}")


def _draw_pyglet_label(label: pyglet.text.Label) -> None:
    """
    Helper for drawing pyglet labels with rotation within arcade.

    Args:
       label: The pyglet label to draw
    """
    assert isinstance(label, pyglet.text.Label)
    label.draw()


class Text:
    """
    An object-oriented way to draw text to the screen.

    .. tip:: Use this class when performance matters!

       Unlike :py:func:`~arcade.draw_text`, this class does not risk
       wasting time recalculating and re-setting any text each time
       :py:meth:`~arcade.Text.draw` is called. This makes it faster
       while:

       - requiring you to manage instances and drawing yourself
       - using negligible extra RAM

       The speed advantage scales as more text needs to be drawn
       to the screen.

    .. tip:: Batch drawing larger amounts of text

       Text objects can also be assigned a pyglet batch for batch
       rendering small or large amounts of text instances. This
       is by far the most efficient way to draw text::

            from pyglet.graphics import Batch

            batch = Batch()
            text_1 = Text("Hello, World 1", 0, 50, batch=batch)
            text_2 = Text("Hello, World 2", 0, 100, batch=batch)
            text_3 = Text("Hello, World 2", 0, 150, batch=batch)
            # Draw the batch
            batch.draw()
            # Remove a text instance from the batch
            text_2.batch = None

       The text instances an also be modified while in the batch
       such as changing the text value, position, or color.

    The constructor arguments work identically to those of
    :py:func:`~arcade.draw_text`. See its documentation for in-depth
    explanation for how to use each of them. For example code, see :ref:`drawing_text_objects`.

    Args:
        text: Initial text to display. Can be an empty string
        x: x position to align the text's anchor point with
        y: y position to align the text's anchor point with
        z: z position to align the text's anchor point with
        color: Color of the text as an RGBA tuple or a
            :py:class:`~arcade.types.Color` instance.
        font_size: Size of the text in points
        width: A width limit in pixels
        align: Horizontal alignment; values other than "left" require width to be set.
            Valid options: ``"left"``, ``"center"``, ``"right"``.
        font_name: A font name, path to a font file, or list of names
        bold: Whether to draw the text as bold, and if a string,
              how bold. See :py:attr:`.bold` to learn more.
        italic: Whether to draw the text as italic
        anchor_x: How to calculate the anchor point's x coordinate.
                  Options: "left", "center", or "right"
        anchor_y: How to calculate the anchor point's y coordinate.
                  Options: "top", "bottom", "center", or "baseline".
        multiline: Requires width to be set; enables word wrap rather than clipping
        rotation: rotation in degrees, clockwise from horizontal
        batch: The batch to add the text to (for batch rendering text)
        group: The specific group in a a batch to add the text to
            (for batch rendering text)

    All constructor arguments other than ``text`` have a corresponding
    property. To access the current text, use the ``value`` property
    instead.

    By default, the text is placed so that:

    - the left edge of its bounding box is at ``x``
    - its baseline is at ``y``

    The baseline is located along the line the bottom of the text would
    be written on, excluding letters with tails such as y:

        .. figure:: ../images/text_anchor_y.png
           :width: 40%

           The blue line is the baseline for the string ``"Python"``

    ``rotation`` allows for the text to be rotated around the anchor
    point by the passed number of degrees. Positive values rotate
    clockwise from horizontal, while negative values rotate
    counter-clockwise:

        .. figure:: ../images/text_rotation_degrees.png
           :width: 55%

           Rotation around the default anchor (
           ``anchor_y="baseline"`` and ``anchor_x="left"``)

    This class is a wrapper around pyglet's :py:class:`pyglet.text.Label`.
    More advanced users can use pyglet's label directly if preferred.
    """

    def __init__(
        self,
        text: str,
        x: float,
        y: float,
        color: RGBOrA255 = arcade.color.WHITE,
        font_size: float = 12,
        width: int | None = None,
        align: str = "left",
        font_name: FontNameOrNames = ("calibri", "arial"),
        bold: bool | str = False,
        italic: bool = False,
        anchor_x: str = "left",
        anchor_y: str = "baseline",
        multiline: bool = False,
        rotation: float = 0,
        batch: pyglet.graphics.Batch | None = None,
        group: pyglet.graphics.Group | None = None,
        z: float = 0,
        **kwargs,
    ):
        self._arguments = dict(
            text=text,
            x=x,
            y=y,
            color=Color.from_iterable(color),
            font_size=font_size,
            width=width,
            align=align,
            font_name=font_name,
            weight=pyglet.text.Weight.BOLD if bold else pyglet.text.Weight.NORMAL,
            italic=italic,
            anchor_x=anchor_x,
            anchor_y=anchor_y,
            multiline=multiline,
            rotation=rotation,
            batch=batch,
            group=group,
            z=z,
            **kwargs,
        )

        if align not in ("left", "center", "right"):
            raise ValueError("The 'align' parameter must be equal to 'left', 'right', or 'center'.")

        if multiline and not width:
            raise ValueError(
                f"The 'width' parameter must be set to a non-zero value when 'multiline' is True, "
                f"but got {width!r}."
            )

        self._initialized = False
        try:
            self._init_deferred()
        except Exception:
            pass

    @property
    def label(self) -> pyglet.text.Label:
        """
        The underlying pyglet.Label instance.
        """
        if not self._initialized:
            self._init_deferred()
        return self._label

    def initialize(self) -> None:
        """
        Manually initialize the Text if it was lazy loaded.
        This has no effect if the Text was already initialized.
        """
        if self._initialized:
            return
        self._init_deferred()

    def _init_deferred(self):
        """
        Deferred initialization when lazy loaded
        """
        # NOTE: Give the user a clear error message stating that the window is not created yet
        arcade.get_window()

        self._arguments["font_name"] = _attempt_font_name_resolution(self._arguments["font_name"])  # type: ignore
        self._label = pyglet.text.Label(**self._arguments)  # type: ignore
        self._initialized = True

    def __enter__(self):
        """
        Update multiple attributes of this text,
        using efficient update mechanism of the underlying ``pyglet.Label``
        """
        self.label.begin_update()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.label.end_update()

    @property
    def batch(self) -> pyglet.graphics.Batch | None:
        """The batch this text is in, if any.

        Can be unset by setting to ``None``.
        """
        return self.label.batch

    @batch.setter
    def batch(self, batch: pyglet.graphics.Batch):
        self.label.batch = batch

    @property
    def group(self) -> pyglet.graphics.Group | None:
        """
        The specific group in a batch the text should belong to.

        This is normally not necessary to specify unless you are
        batching very large sets of text needing to separate into
        groups or even mix with other pyglet batch content.
        """
        return self.label.group

    @group.setter
    def group(self, group: pyglet.graphics.Group):
        self.label.group = group

    @property
    def value(self) -> str:
        """
        Get or set the current text string to display.

        The value assigned will be converted to a string.
        """
        return self.label.text

    @value.setter
    def value(self, value: Any):
        value = str(value)
        if self.label.text == value:
            return
        self.label.text = value

    @property
    def text(self) -> str:
        """
        Get or set the current text string to display.

        The value assigned will be converted to a string.

        This is an alias for :py:attr:`~arcade.Text.value`
        """
        return self.label.text

    @text.setter
    def text(self, value: Any):
        value = str(value)
        if self.label.text == value:
            return
        self.label.text = value

    @property
    def x(self) -> float:
        """Get or set the x position of the label."""
        return self.label.x

    @x.setter
    def x(self, x: float) -> None:
        if self.label.x == x:
            return
        self.label.x = x

    @property
    def y(self) -> float:
        """Get or set the y position of the label."""
        return self.label.y

    @y.setter
    def y(self, y: float):
        if self.label.y == y:
            return
        self.label.y = y

    @property
    def z(self) -> float:
        """Get or set the z position of the label."""
        return self.label.z

    @z.setter
    def z(self, z: float):
        if self.label.z == z:
            return
        self.label.z = z

    @property
    def font_name(self) -> FontNameOrNames:
        """Get or set the font name(s) for the label."""
        if not isinstance(self.label.font_name, str):
            return tuple(self.label.font_name)
        else:
            return self.label.font_name

    @font_name.setter
    def font_name(self, font_name: FontNameOrNames) -> None:
        if isinstance(font_name, str):
            self.label.font_name = font_name
        else:
            self.label.font_name = list(font_name)

    @property
    def font_size(self) -> float:
        """Get or set the font size of the label."""
        return self.label.font_size

    @font_size.setter
    def font_size(self, font_size: float):
        self.label.font_size = font_size

    @property
    def anchor_x(self) -> str:
        """
        Get or set the horizontal anchor.

        Options: ``"left"``, ``"center"``, or ``"right"``
        """
        return self.label.anchor_x

    @anchor_x.setter
    def anchor_x(self, anchor_x: str):
        self.label.anchor_x = anchor_x  # type: ignore

    @property
    def anchor_y(self) -> str:
        """
        Get or set the vertical anchor.

        Options : ``"top"``, ``"bottom"``, ``"center"``, or ``"baseline"``
        """
        return self.label.anchor_y

    @anchor_y.setter
    def anchor_y(self, anchor_y: str):
        self.label.anchor_y = anchor_y  # type: ignore

    @property
    def rotation(self) -> float:
        """Get or set the clockwise rotation"""
        return self.label.rotation

    @rotation.setter
    def rotation(self, rotation: float):
        self.label.rotation = rotation

    @property
    def color(self) -> Color:
        """Get or set the text color for the label."""
        return Color.from_iterable(self.label.color)

    @color.setter
    def color(self, color: RGBOrA255):
        self.label.color = Color.from_iterable(color)

    @property
    def width(self) -> int | None:
        """
        Get or set the width of the label in pixels.

        This value affects text flow when multiline text is used.
        If you are looking for the physical size if the text, see
        :py:attr:`~arcade.Text.content_width`
        """
        return self.label.width

    @width.setter
    def width(self, width: int):
        self.label.width = width

    @property
    def height(self) -> int | None:
        """
        Get or set the height of the label in pixels.

        This value affects text flow when multiline text is used.
        If you are looking for the physical size if the text, see
        :py:attr:`~arcade.Text.content_height`
        """
        return self.label.height

    @height.setter
    def height(self, value: int):
        self.label.height = value

    @property
    def size(self):
        """Get the size of the label."""
        return self.label.width, self.label.height

    @property
    def content_width(self) -> int:
        """Get the pixel width of the text contents."""
        return self.label.content_width

    @property
    def content_height(self) -> int:
        """Get the pixel height of the text content."""
        return self.label.content_height

    @property
    def left(self) -> float:
        """Pixel location of the left content border."""
        return self.label.left

    @property
    def right(self) -> float:
        """Pixel location of the right content border."""
        return self.label.right

    @property
    def top(self) -> float:
        """Pixel location of the top content border."""
        return self.label.top

    @property
    def bottom(self) -> float:
        """Pixel location of the bottom content border."""
        return self.label.bottom

    @property
    def content_size(self) -> tuple[int, int]:
        """Get the pixel width and height of the text contents."""
        return self.label.content_width, self.label.content_height

    @property
    def align(self) -> str:
        """Horizontal alignment; values other than ``"left"`` require width to be set.

        Valid options: ``"left"``, ``"center"``, ``"right"``.
        """
        return self.label.get_style("align")  # type: ignore

    @align.setter
    def align(self, align: str):
        self.label.set_style("align", align)

    @property
    def bold(self) -> bool | str:
        """
        Get or set bold state of the label.

        The supported values include:

        * ``"black"``
        * ``"bold" (same as ``True``)
        * ``"semibold"``
        * ``"semilight"``
        * ``"light"``

        """
        return self.label.weight == pyglet.text.Weight.BOLD

    @bold.setter
    def bold(self, bold: bool | str):
        self.label.weight = pyglet.text.Weight.BOLD if bold else pyglet.text.Weight.NORMAL

    @property
    def italic(self) -> bool | str:
        """Get or set the italic state of the label."""
        return self.label.italic

    @italic.setter
    def italic(self, italic: bool | str):
        self.label.italic = italic

    @property
    def multiline(self) -> bool:
        """Get or set the multiline flag of the label."""
        return self.label.multiline

    @multiline.setter
    def multiline(self, multiline: bool):
        self.label.multiline = multiline

    def draw(self) -> None:
        """
        Draw the label to the screen at its current ``x`` and ``y`` position.

        .. warning:: Cameras affect text drawing!

            If you want to draw a custom GUI that doesn't move with the
            game world, you will need a second :py:class:`~arcade.Camera`
            instance. For information on how to do this, see
            :ref:`sprite_move_scrolling`.
        """
        _draw_pyglet_label(self.label)

    def draw_debug(
        self,
        anchor_color: RGBOrA255 = arcade.color.RED,
        background_color: RGBOrA255 = arcade.color.DARK_BLUE,
        outline_color: RGBOrA255 = arcade.color.WHITE,
    ) -> None:
        """
        Draw test with debug geometry showing the content
        area, outline and the anchor point.

        Args:
            anchor_color: Color of the anchor point
            background_color: Color the content background
            outline_color: Color of the content outline
        """
        left = self.left
        right = self.right
        top = self.top
        bottom = self.bottom

        # Draw background
        arcade.draw_lrbt_rectangle_filled(left, right, bottom, top, color=background_color)

        # Draw outline
        arcade.draw_lrbt_rectangle_outline(left, right, bottom, top, color=outline_color)

        # Draw anchor
        arcade.draw_point(self.x, self.y, color=anchor_color, size=6)

        _draw_pyglet_label(self.label)

    @property
    def position(self) -> Point:
        """
        The current x, y position as a tuple.

        This is faster than setting x and y position separately
        because the underlying geometry only needs to change position once.
        """
        return self.label.x, self.label.y

    @position.setter
    def position(self, point: Point):
        # Starting with Pyglet 2.0b2 label positions take a z parameter.
        x, y, *z = point

        if z:
            self.label.position = x, y, z[0]
        else:
            self.label.position = x, y, self.label.z

    @property
    def tracking(self) -> float | None:
        """
        Get/set the tracking amount for this text object, or rather,
        the added space between each character.

        The value is an amount in pixels and can be negative.
        To convert from the em unit, use Text.em_to_px().

        Returns:
            a pixel amount, or None if the tracking is inconsistent.
        """
        kerning = self.label.get_style("kerning")
        return kerning if kerning != pyglet.text.document.STYLE_INDETERMINATE else None

    @tracking.setter
    def tracking(self, value: float):
        self.label.set_style("kerning", value)

    def em_to_px(self, em: float) -> float:
        """Convert from an em value to a pixel amount.

        1em is defined as ``font_size`` pt.
        """
        return (em * self.font_size) * (4 / 3)

    def px_to_em(self, px: float) -> float:
        """Convert from a pixel amount to a value in ems.

        1em is defined as ``font_size`` pt.
        """
        return px / (4 / 3) / self.font_size


def create_text_sprite(
    text: str,
    color: RGBOrA255 = arcade.color.WHITE,
    font_size: float = 12.0,
    width: int | None = None,
    align: str = "left",
    font_name: FontNameOrNames = ("calibri", "arial"),
    bold: bool | str = False,
    italic: bool = False,
    anchor_x: str = "left",
    multiline: bool = False,
    texture_atlas: TextureAtlasBase | None = None,
    background_color: RGBOrA255 | None = None,
) -> arcade.Sprite:
    """
    Creates a sprite containing text based off of :py:class:`~arcade.Text`.

    Internally this creates a Text object and an empty texture. It then uses either the
    provided texture atlas, or gets the default one, and draws the Text object into the
    texture atlas.

    It then creates a sprite referencing the newly created texture, and positions it
    accordingly, and that is final result that is returned from the function.

    If you are providing a custom texture atlas, something important to keep in mind is
    that the resulting Sprite can only be added to SpriteLists which use that atlas. If
    it is added to a SpriteList which uses a different atlas, you will likely just see
    a black box drawn in its place.

    Args:
        text: Initial text to display. Can be an empty string
        color: Color of the text as an RGBA tuple or a
            :py:class:`~arcade.types.Color` instance.
        font_size: Size of the text in points
        width: A width limit in pixels
        align: Horizontal alignment; values other than "left" require width to be set.
            Valid options: ``"left"``, ``"center"``, ``"right"``.
        font_name: A font name, path to a font file, or list of names
        bold: Whether to draw the text as bold, and if a string,
              how bold. See :py:attr:`arcade.gui.widgets.text.bold` to learn more.
        italic: Whether to draw the text as italic
        anchor_x: How to calculate the anchor point's x coordinate.
                  Options: "left", "center", or "right"
        multiline: Requires width to be set; enables word wrap rather than clipping
        background_color: The background color of the text. If None, the background
            will be transparent.
        texture_atlas: The texture atlas to use for the
            newly created texture. The default global atlas will be used if this is None.
    """
    text_object = Text(
        text,
        x=0,
        y=0,
        color=color,
        font_size=font_size,
        width=width,
        align=align,
        font_name=font_name,
        bold=bold,
        italic=italic,
        anchor_x=anchor_x,
        anchor_y="baseline",
        multiline=multiline,
    )

    size = (
        int(text_object.right - text_object.left),
        int(text_object.top - text_object.bottom),
    )
    text_object.y = -text_object.bottom
    texture = arcade.Texture.create_empty(text, size)

    if not texture_atlas:
        texture_atlas = arcade.get_window().ctx.default_atlas
    texture_atlas.add(texture)
    with texture_atlas.render_into(texture) as fbo:
        fbo.clear(color=background_color or arcade.color.TRANSPARENT_BLACK)
        text_object.draw()

    return arcade.Sprite(
        texture,
        center_x=text_object.right - (size[0] / 2),
        center_y=text_object.top,
    )


@warning(
    message=(
        "draw_text is an extremely slow function for displaying text. "
        "Consider using Text objects instead."
    ),
    warning_type=PerformanceWarning,
)
def draw_text(
    text: Any,
    x: float,
    y: float,
    color: RGBOrA255 = arcade.color.WHITE,
    font_size: float = 12.0,
    width: int | None = None,
    align: str = "left",
    font_name: FontNameOrNames = ("calibri", "arial"),
    bold: bool | str = False,
    italic: bool = False,
    anchor_x: str = "left",
    anchor_y: str = "baseline",
    multiline: bool = False,
    rotation: float = 0,
    z: float = 0,
):
    """
    A simple way for beginners to draw text.

    .. warning:: Use :py:class:`arcade.Text` objects instead.

        This method of drawing text is very slow
        and might be removed in the near future.
        Text objects can be 10-100 times faster
        depending on the use case.

    .. warning:: Cameras affect text drawing!

        If you want to draw a custom GUI that doesn't move with the
        game world, you will need a second camera. For information on
        how to do this, see :ref:`sprite_move_scrolling`.

    This function lets you start draw text easily with better
    performance than the old pillow-based text. If you need even higher
    performance, consider using :py:class:`~arcade.Text`.

    Example code can be found at :ref:`drawing_text`.

    Args:
        text: Initial text to display. Can be an empty string
        x: x position to align the text's anchor point with
        y: y position to align the text's anchor point with
        z: z position to align the text's anchor point with
        color: Color of the text as an RGBA tuple or a
            :py:class:`~arcade.types.Color` instance.
        font_size: Size of the text in points
        width: A width limit in pixels
        align: Horizontal alignment; values other than "left" require width to be set.
            Valid options: ``"left"``, ``"center"``, ``"right"``.
        font_name: A font name, path to a font file, or list of names
        bold: Whether to draw the text as bold, and if a string,
              how bold. See :py:attr:`arcade.gui.widgets.text.bold` to learn more.
        italic: Whether to draw the text as italic
        anchor_x: How to calculate the anchor point's x coordinate.
                  Options: "left", "center", or "right"
        anchor_y: How to calculate the anchor point's y coordinate.
                  Options: "top", "bottom", "center", or "baseline".
        multiline: Requires width to be set; enables word wrap rather than clipping
        rotation: rotation in degrees, clockwise from horizontal

    By default, the text is placed so that:

    - the left edge of its bounding box is at ``x``
    - its baseline is at ``y``

    The baseline of text is the line it would be written on:

        .. figure:: ../images/text_anchor_y.png
           :width: 40%

           The blue line is the baseline for the string ``"Python"``

    ``font_name`` can be any of the following:

    - a built-in font in the :ref:`Resources`
    - the name of a system font
    - a path to a font on the system
    - a `tuple` containing any mix of the previous three

    Each entry provided will be tried in order until one is found. If
    none of the fonts are found, a default font will be chosen (usually
    Arial).

    ``anchor_x`` and ``anchor_y`` specify how to calculate the anchor point,
    which affects how the text is:

    - Placed relative to ``x`` and ``y``
    - Rotated

    By default, the text is drawn so that ``x`` is at the left of
    the text's bounding box and ``y`` is at the baseline.

    You can set a custom anchor point by passing combinations of the
    following values for ``anchor_x`` and ``anchor_y``:

    .. list-table:: Values allowed by ``anchor_x``
        :widths: 20 40 40
        :header-rows: 1

        * - String value
          - Practical Effect
          - Anchor Position

        * - ``"left"`` `(default)`
          - Text drawn with its left side at ``x``
          - Anchor point on the left side of the text's bounding box

        * - ``"center"``
          - Text drawn horizontally centered on ``x``
          - Anchor point at horizontal center of text's bounding box

        * - ``"right"``
          - Text drawn with its right side at ``x``
          - Anchor placed on the right side of the text's bounding box


    .. list-table:: Values allowed by ``anchor_y``
        :widths: 20 40 40
        :header-rows: 1

        * - String value
          - Practical Effect
          - Anchor Position

        * - ``"baseline"`` `(default)`
          - Text drawn with baseline on ``y``.
          - Anchor placed at the text rendering baseline

        * - ``"top"``
          - Text drawn with its top aligned with ``y``
          - Anchor point placed at the top of the text

        * - ``"bottom"``
          - Text drawn with its absolute bottom aligned with ``y``,
            including the space for tails on letters such as y and g
          - Anchor point placed at the bottom of the text after the
            space allotted for letters such as y and g

        * - ``"center"``
          - Text drawn with its vertical center on ``y``
          - Anchor placed at the vertical center of the text


    ``rotation`` allows for the text to be rotated around the anchor
    point by the passed number of degrees. Positive values rotate
    clockwise from horizontal, while negative values rotate
    counter-clockwise:

        .. figure:: ../images/text_rotation_degrees.png
           :width: 55%

           Rotation around the default anchor point (
           ``anchor_y="baseline"`` and ``anchor_x="left"``)


    It can be helpful to think of this function working as follows:

    1. Text layout and alignment are calculated:

        1. The text's characters are laid out within a bounding box
           according to the current styling options

        2. The anchor point on the text is calculated based on
           the text value, styling, as well as values for ``anchor_x``
           and ``anchor_y``

    2. The text is placed so its anchor point is at ``(x,
       y))``

    3. The text is rotated around its anchor point before finally
       being drawn

    This function is less efficient than using :py:class:`~arcade.Text`
    because some steps above can be repeated each time a call is
    made rather than fully cached as with the class.

    """
    # See : https://github.com/pyglet/pyglet/blob/ff30eadc2942553c9de96d6ce564ad1bc3128fb4/pyglet/text/__init__.py#L401

    color = Color.from_iterable(color)
    # Cache the states that are expensive to change
    key = f"{font_size}{font_name}{bold}{italic}{anchor_x}{anchor_y}{align}{width}{rotation}"
    ctx = arcade.get_window().ctx
    label = ctx.label_cache.get(key)

    if align not in ("left", "center", "right"):
        raise ValueError("The 'align' parameter must be equal to 'left', 'right', or 'center'.")

    if multiline and not width:
        raise ValueError(
            f"The 'width' parameter must be set to a non-zero value when 'multiline' is True, "
            f"but got {width!r}."
        )

    if not label:
        adjusted_font = _attempt_font_name_resolution(font_name)

        label = arcade.Text(
            text=str(text),
            x=x,
            y=y,
            z=z,
            font_name=adjusted_font,
            font_size=font_size,
            anchor_x=anchor_x,
            anchor_y=anchor_y,
            color=color,
            width=width,
            align=align,
            bold=bold,
            italic=italic,
            multiline=multiline,
            rotation=rotation,
        )
        ctx.label_cache[key] = label

    # These updates are quite expensive
    if label.text != text:
        label.text = str(text)
    if label.x != x or label.y != y or label.z != z:
        label.position = x, y, z  # type: ignore
    if label.color != color:
        label.color = color
    if label.rotation != rotation:
        label.rotation = rotation

    label.draw()
    # This is absolutely necessary to prevent the vertex buffers
    # to be altered while another one is drawing. If the same cached
    # label is used multiple times in a single frame it's a disaster.
    ctx.flush()



"""
Platformer Game

python -m arcade.examples.platform_tutorial.15_ladders_moving_platforms
"""
import arcade

# Constants
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Platformer"

# Constants used to scale our sprites from their original size
TILE_SCALING = 0.5
COIN_SCALING = 0.5

# Movement speed of player, in pixels per frame
PLAYER_MOVEMENT_SPEED = 5
GRAVITY = 1
PLAYER_JUMP_SPEED = 20


class GameView(arcade.Window):
    """
    Main application class.
    """

    def __init__(self):

        # Call the parent class and set up the window
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

        # Variable to hold our texture for our player
        self.player_texture = None

        # Separate variable that holds the player sprite
        self.player_sprite = None

        # Variable to hold our Tiled Map
        self.tile_map = None

        # Replacing all of our SpriteLists with a Scene variable
        self.scene = None

        # A variable to store our camera object
        self.camera = None

        # A variable to store our gui camera object
        self.gui_camera = None

        # This variable will store our score as an integer.
        self.score = 0

        # This variable will store the text for score that we will draw to the screen.
        self.score_text = None

        # Where is the right edge of the map?
        self.end_of_map = 0

        # Should we reset the score?
        self.reset_score = True

        # Load sounds
        self.collect_coin_sound = arcade.load_sound(":resources:sounds/coin1.wav")
        self.jump_sound = arcade.load_sound(":resources:sounds/jump1.wav")
        self.gameover_sound = arcade.load_sound(":resources:sounds/gameover1.wav")

    def setup(self):
        """Set up the game here. Call this function to restart the game."""
        layer_options = {
            "Platforms": {
                "use_spatial_hash": True
            },
            "Moving Platforms": {
                "use_spatial_hash": False
            },
            "Ladders": {
                "use_spatial_hash": True
            }
        }

        # Load our TileMap
        self.tile_map = arcade.load_tilemap(
            ":resources:tiled_maps/map_with_ladders.json",
            scaling=TILE_SCALING,
            layer_options=layer_options,
        )

        # Create our Scene Based on the TileMap
        self.scene = arcade.Scene.from_tilemap(self.tile_map)

        self.player_texture = arcade.load_texture(
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_idle.png"
        )

        self.player_sprite = arcade.Sprite(self.player_texture)
        self.player_sprite.center_x = 128
        self.player_sprite.center_y = 128
        self.scene.add_sprite("Player", self.player_sprite)

        # Create a Platformer Physics Engine, this will handle moving our
        # player as well as collisions between the player sprite and
        # whatever SpriteList we specify for the walls.
        # It is important to supply static to the walls parameter. There is a
        # platforms parameter that is intended for moving platforms.
        # If a platform is supposed to move, and is added to the walls list,
        # it will not be moved.
        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite,
            walls=self.scene["Platforms"],
            gravity_constant=GRAVITY,
            platforms=self.scene["Moving Platforms"],
            ladders=self.scene["Ladders"]
        )

        # Initialize our camera, setting a viewport the size of our window.
        self.camera = arcade.Camera2D()

        # Initialize our gui camera, initial settings are the same as our world camera.
        self.gui_camera = arcade.Camera2D()

        # Reset the score if we should
        if self.reset_score:
            self.score = 0
        self.reset_score = True

        # Initialize our arcade.Text object for score
        self.score_text = arcade.Text(f"Score: {self.score}", x=0, y=5)

        self.background_color = arcade.csscolor.CORNFLOWER_BLUE

        # Calculate the right edge of the map in pixels
        self.end_of_map = (self.tile_map.width * self.tile_map.tile_width)
        self.end_of_map *= self.tile_map.scaling

    def on_draw(self):
        """Render the screen."""

        # Clear the screen to the background color
        self.clear()

        # Activate our camera before drawing
        self.camera.use()

        # Draw our Scene
        self.scene.draw()

        # Activate our GUI camera
        self.gui_camera.use()

        # Draw our Score
        self.score_text.draw()

    def on_update(self, delta_time):
        """Movement and Game Logic"""

        # Move the player using our physics engine
        self.physics_engine.update()

        # See if we hit any coins
        coin_hit_list = arcade.check_for_collision_with_list(
            self.player_sprite, self.scene["Coins"]
        )

        # Loop through each coin we hit (if any) and remove it
        for coin in coin_hit_list:
            # Remove the coin
            coin.remove_from_sprite_lists()
            arcade.play_sound(self.collect_coin_sound)
            self.score += 75
            self.score_text.text = f"Score: {self.score}"

        # Center our camera on the player
        self.camera.position = self.player_sprite.position

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed."""

        if key == arcade.key.ESCAPE:
            self.setup()

        if key == arcade.key.UP or key == arcade.key.W:
            if self.physics_engine.is_on_ladder():
                self.player_sprite.change_y = PLAYER_MOVEMENT_SPEED
            elif self.physics_engine.can_jump():
                self.player_sprite.change_y = PLAYER_JUMP_SPEED
                arcade.play_sound(self.jump_sound)

        if key == arcade.key.DOWN or key == arcade.key.S:
            if self.physics_engine.is_on_ladder():
                self.player_sprite.change_y = -PLAYER_MOVEMENT_SPEED

        if key == arcade.key.LEFT or key == arcade.key.A:
            self.player_sprite.change_x = -PLAYER_MOVEMENT_SPEED
        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.player_sprite.change_x = PLAYER_MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """Called whenever a key is released."""

        if key == arcade.key.LEFT or key == arcade.key.A:
            self.player_sprite.change_x = 0
        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.player_sprite.change_x = 0
        elif key == arcade.key.UP or key == arcade.key.W:
            self.player_sprite.change_y = 0
        elif key == arcade.key.DOWN or key == arcade.key.S:
            self.player_sprite.change_y = 0


def main():
    """Main function"""
    window = GameView()
    window.setup()
    arcade.run()


if __name__ == "__main__":
    main()



"""
Physics engines for top-down or platformers.
"""

import math
from collections.abc import Iterable

from arcade import (
    BasicSprite,
    Sprite,
    SpriteSequence,
    SpriteType,
    check_for_collision,
    check_for_collision_with_lists,
)
from arcade.math import get_distance

__all__ = ["PhysicsEngineSimple", "PhysicsEnginePlatformer"]

from arcade.utils import Chain, copy_dunders_unimplemented


def _wiggle_until_free(colliding: Sprite, walls: Iterable[SpriteSequence[BasicSprite]]) -> None:
    """Kludge to 'guess' a colliding sprite out of a collision.

    It works by iterating over increasing wiggle sizes of 8 points
    around the ``colliding`` sprite's original center position. Each
    time it fails to find a free position. Although the wiggle distance
    starts at 1, it grows quickly since each failed iteration multiplies
    wiggle distance by two.

    Args:
        colliding:
            A sprite to move out of the given list of SpriteLists.
        walls:
            A list of walls to guess our way out of.
    """
    # Original x & y of the moving object
    o_x, o_y = colliding.position

    # fmt: off
    try_list = [  # Allocate once so we don't recreate or gc
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0,
    ]

    wiggle_distance = 1
    while True:
        # Cache our variant dimensions
        o_x_plus  = o_x + wiggle_distance
        o_y_plus  = o_y + wiggle_distance
        o_x_minus = o_x - wiggle_distance
        o_y_minus = o_y - wiggle_distance

        # Burst setting of no-gc region is cheaper than nested lists
        try_list[:] = (
            o_x       , o_y_plus ,
            o_x       , o_y_minus,
            o_x_plus  , o_y      ,
            o_x_minus , o_y      ,
            o_x_plus  , o_y_plus ,
            o_x_plus  , o_y_minus,
            o_x_minus , o_y_plus ,
            o_x_minus , o_y_minus
        )
        # fmt: on

        # Iterate and slice the try_list
        for strided_index in range(0, 16, 2):
            x, y = try_list[strided_index:strided_index + 2]
            colliding.position = x, y
            check_hit_list = check_for_collision_with_lists(colliding, walls)
            # print(f"Vary {vary} ({trapped.center_x} {trapped.center_y}) "
            #       f"= {len(check_hit_list)}")
            if len(check_hit_list) == 0:
                return
        wiggle_distance *= 2


def _move_sprite(
    moving_sprite: Sprite, can_collide: Iterable[SpriteSequence[SpriteType]], ramp_up: bool
) -> list[SpriteType]:
    """Update a sprite's angle and position, returning a list of collisions.

    The steps covered are:

    #. Rotation
    #. Move in the y direction
    #. Move in the x direction

    Args:
        moving_sprite:
            The sprite to move.
        can_collide:
            An iterable source of SpriteList objects which can be
            collided with.
        ramp_up:
            Whether to enable platformer-like ramp support for x
            direction movement.
    Returns:
        A list of other individual sprites the ``moving_sprite``
        collided with.
    """
    # See if we are starting this turn with a sprite already colliding with us.
    if len(check_for_collision_with_lists(moving_sprite, can_collide)) > 0:
        _wiggle_until_free(moving_sprite, can_collide)

    original_x, original_y = moving_sprite.position
    original_angle = moving_sprite.angle

    # --- Rotate
    rotating_hit_list = []
    if moving_sprite.change_angle:
        # Rotate
        moving_sprite.angle += moving_sprite.change_angle

        # Resolve collisions caused by rotating
        rotating_hit_list = check_for_collision_with_lists(moving_sprite, can_collide)

        if len(rotating_hit_list) > 0:
            max_distance = (moving_sprite.width + moving_sprite.height) / 2

            # Resolve any collisions by this weird kludge
            _wiggle_until_free(moving_sprite, can_collide)
            if (
                get_distance(original_x, original_y, moving_sprite.center_x, moving_sprite.center_y)
                > max_distance
            ):
                # Ok, glitched trying to rotate. Reset.
                moving_sprite.position = original_x, original_y
                moving_sprite.angle = original_angle

    # --- Move in the y direction
    moving_sprite.center_y += moving_sprite.change_y

    # Check for wall hit
    hit_list_x = check_for_collision_with_lists(moving_sprite, can_collide)
    # print(f"Post-y move {hit_list_x}")
    complete_hit_list = hit_list_x

    # If we hit a wall, move so the edges are at the same point
    if len(hit_list_x) > 0:
        if moving_sprite.change_y > 0:
            while len(check_for_collision_with_lists(moving_sprite, can_collide)) > 0:
                moving_sprite.center_y -= 1
            # print(f"Spot X ({self.player_sprite.center_x}, {self.player_sprite.center_y})"
            #       f" {self.player_sprite.change_y}")
        elif moving_sprite.change_y < 0:
            # Reset number of jumps
            for item in hit_list_x:
                while check_for_collision(moving_sprite, item):
                    # self.player_sprite.bottom = item.top <- Doesn't work for ramps
                    moving_sprite.center_y += 0.25

                # NOTE: Not all sprites have velocity
                if getattr(item, "change_x", 0.0) != 0:
                    moving_sprite.center_x += item.change_x  # type: ignore

            # print(f"Spot Y ({self.player_sprite.center_x}, {self.player_sprite.center_y})")
        else:
            pass
            # TODO: The code below can't execute, as "item" doesn't
            # exist. In theory, this condition should never be arrived at.
            # Collision while player wasn't moving, most likely
            # moving platform.
            # if self.player_sprite.center_y >= item.center_y:
            #     self.player_sprite.bottom = item.top
            # else:
            #     self.player_sprite.top = item.bottom
        moving_sprite.change_y = min(0.0, getattr(hit_list_x[0], "change_y", 0.0))

    # print(f"Spot D ({self.player_sprite.center_x}, {self.player_sprite.center_y})")
    moving_sprite.center_y = round(moving_sprite.center_y, 2)
    # print(f"Spot Q ({self.player_sprite.center_x}, {self.player_sprite.center_y})")

    # end_time = time.time()
    # print(f"Move 1 - {end_time - start_time:7.4f}")
    # start_time = time.time()

    loop_count = 0
    # --- Move in the x direction
    if moving_sprite.change_x:
        # Keep track of our current y, used in ramping up
        almost_original_y = moving_sprite.center_y

        # Strip off sign so we only have to write one version of this for
        # both directions
        direction = math.copysign(1, moving_sprite.change_x)
        cur_x_change = abs(moving_sprite.change_x)
        upper_bound = cur_x_change
        lower_bound: float = 0
        cur_y_change: float = 0

        exit_loop = False
        while not exit_loop:
            loop_count += 1
            # print(f"{cur_x_change=}, {upper_bound=}, {lower_bound=}, {loop_count=}")

            # Move sprite and check for collisions
            moving_sprite.center_x = original_x + cur_x_change * direction
            collision_check = check_for_collision_with_lists(moving_sprite, can_collide)

            # Update collision list
            for sprite in collision_check:
                if sprite not in complete_hit_list:
                    complete_hit_list.append(sprite)

            # Did we collide?
            if len(collision_check) > 0:
                # We did collide. Can we ramp up and not collide?
                if ramp_up:
                    cur_y_change = cur_x_change
                    moving_sprite.center_y = original_y + cur_y_change

                    collision_check = check_for_collision_with_lists(moving_sprite, can_collide)
                    if len(collision_check) > 0:
                        cur_y_change -= cur_x_change
                    else:
                        while (len(collision_check) == 0) and cur_y_change > 0:
                            # print("Ramp up check")
                            cur_y_change -= 1
                            moving_sprite.center_y = almost_original_y + cur_y_change
                            collision_check = check_for_collision_with_lists(
                                moving_sprite, can_collide
                            )
                        cur_y_change += 1
                        collision_check = []

                if len(collision_check) > 0:
                    # print(f"Yes @ {cur_x_change}")
                    upper_bound = cur_x_change - 1
                    if upper_bound - lower_bound <= 0:
                        cur_x_change = lower_bound
                        exit_loop = True
                        # print(f"Exit 2 @ {cur_x_change}")
                    else:
                        cur_x_change = (upper_bound + lower_bound) // 2
                else:
                    exit_loop = True
                    # print(f"Exit 1 @ {cur_x_change}")

            else:
                # No collision. Keep this new position and exit
                lower_bound = cur_x_change
                if upper_bound - lower_bound <= 0:
                    # print(f"Exit 3 @ {cur_x_change}")
                    exit_loop = True
                else:
                    # print(f"No @ {cur_x_change}")
                    cur_x_change = (upper_bound + lower_bound) // 2 + (
                        upper_bound + lower_bound
                    ) % 2

        # print(cur_x_change * direction, cur_y_change)
        moved_x = original_x + cur_x_change * direction
        moved_y = almost_original_y + cur_y_change
        moving_sprite.position = moved_x, moved_y
        # print(
        #     f"({moving_sprite.center_x}, {moving_sprite.center_y}) "
        #     f"{cur_x_change * direction}, {cur_y_change}"
        # )

    # Add in rotating hit list
    for sprite in rotating_hit_list:
        if sprite not in complete_hit_list:
            complete_hit_list.append(sprite)

    # end_time = time.time()
    # print(f"Move 2 - {end_time - start_time:7.4f} {loop_count}")

    return complete_hit_list


def _add_to_list(
    dest: list[SpriteSequence[SpriteType]],
    source: SpriteSequence[SpriteType] | Iterable[SpriteSequence[SpriteType]] | None,
) -> None:
    """Helper function to add a SpriteSequence or list of SpriteSequences to a list."""
    if not source:
        return
    elif isinstance(source, SpriteSequence):
        dest.append(source)
    else:
        dest.extend(source)


@copy_dunders_unimplemented
class PhysicsEngineSimple:
    """A basic physics engine best for single-player top-down games.

    This is the easiest engine to get started with. It's best when:

    * You need a top-down view
    * You need to collide with non-moving terrain
    * You don't need anything else

    For side-scrolling games focused on jumping puzzles, you may want
    the :py:class:`PhysicsEnginePlatformer` instead. Experienced users
    may want to try the
    :py:class:`~arcade.pymunk_physics_engine.PymunkPhysicsEngine`.

    Args:
        player_sprite:
            A sprite which will be controlled by the player.

        walls:
            A :py:class:`.SpriteList` or :py:class:`list` of them which
            should stop player movement.
    """

    def __init__(
        self,
        player_sprite: Sprite,
        walls: SpriteSequence[BasicSprite] | Iterable[SpriteSequence[BasicSprite]] | None = None,
    ) -> None:
        self.player_sprite: Sprite = player_sprite
        """The player-controlled :py:class:`.Sprite`."""
        self._walls: list[SpriteSequence[BasicSprite]] = []

        if walls:
            _add_to_list(self._walls, walls)

    @property
    def walls(self) -> list[SpriteSequence[BasicSprite]]:
        """Which :py:class:`.SpriteList` instances block player movement.

        .. important:: Avoid moving sprites in these lists!

                       Doing so incurs performance costs.

        See :py:class:`PhysicsEnginePlatformer.walls` for further information.
        For platformer physics such as moving platforms and gravity, consider
        using the :py:class:`PhysicsEnginePlatformer`.
        """
        return self._walls

    @walls.setter
    def walls(
        self,
        walls: SpriteSequence[BasicSprite] | Iterable[SpriteSequence[BasicSprite]] | None = None,
    ) -> None:
        if walls:
            _add_to_list(self._walls, walls)
        else:
            self._walls.clear()

    @walls.deleter
    def walls(self) -> None:
        self._walls.clear()

    def update(self) -> list[BasicSprite]:
        """Move :py:attr:`player_sprite` and return any colliding sprites.

        Returns:
            A :py:class:`list` of the colliding sprites. If there were
            zero collisions, it will be empty.
        """

        return _move_sprite(self.player_sprite, self.walls, ramp_up=False)


@copy_dunders_unimplemented
class PhysicsEnginePlatformer:
    """A single-player engine with gravity and moving platform support.

    .. _Super Mario Bros.: https://en.wikipedia.org/wiki/Super_Mario_Bros.
    .. _Rayman: https://en.wikipedia.org/wiki/Rayman_(video_game)

    This engine is best for simple versions of platformer games like
    the `Super Mario Bros.`_ (the first Mario game) and `Rayman`_. It
    is more important to pay attention to the performance tips with
    this engine than with  :py:class:`PhysicsEngineSimple`.

    .. important:: For best performance, you must put your sprites
                   in the right group!

    Be sure to add each :py:class:`.Sprite` and :py:class:`.SpriteList`
    to the right group, regardless of whether you do so via arguments
    or properties:

    .. list-table::
       :header-rows: 1

       * - Creation argument
         - Purpose
         - :py:class:`list` property

       * - ``walls``
         - Non-moving sprites the player can stand on.
         - :py:attr:`walls`

       * - ``platforms``
         - Sprites the player can stand on, but which can still move.
         - :py:attr:`platforms`

       * - ``ladders``
         - Ladders which allow gravity-free movement while touched by
           the :py:attr:`player_sprite`.
         - :py:attr:`ladders`

    To learn about the automatic moving platform feature, please see
    :py:attr:`platforms`.

    Not that if you use the :py:class:`list` properties above, you can
    add or remove :py:attr:`platforms` in response to game events. It
    is also possible to add new sprites such as terrain in response to
    gameplay events, but there may be performance implications due to
    the way :py:class:`SpriteList` handles spatial hashing. To learn
    more, see :py:attr:`walls`.

    Args:
        player_sprite:
            The player character's sprite. It will be stored on the engine
            as :py:attr:`player_sprite`.
        platforms:
            The initial list of :py:attr:`platforms`, sprites which can move
            freely.
        gravity_constant:
            A constant to subtract from the :py:attr:`player_sprite`'s
            velocity (:py:attr:`.Sprite.change_y`) each :py:meth:`update`
            when in the air. See :py:attr:`gravity_constant` to learn
            more.
        ladders:
            :py:class:`.Sprite` instances the player can climb without
            being affected by gravity.
        walls:
            :py:class:`Sprite` instances which are static and never move.
            **Do not put moving sprites into this!** See :py:attr:`walls`
            to learn more.

    """

    def __init__(
        self,
        player_sprite: Sprite,
        platforms: SpriteSequence[Sprite] | Iterable[SpriteSequence[Sprite]] | None = None,
        gravity_constant: float = 0.5,
        ladders: SpriteSequence[BasicSprite] | Iterable[SpriteSequence[BasicSprite]] | None = None,
        walls: SpriteSequence[BasicSprite] | Iterable[SpriteSequence[BasicSprite]] | None = None,
    ) -> None:
        if not isinstance(player_sprite, Sprite):
            raise TypeError("player_sprite must be a Sprite, not a basic_sprite!")

        self._ladders: list[SpriteSequence[BasicSprite]] = []
        self._platforms: list[SpriteSequence[Sprite]] = []
        self._walls: list[SpriteSequence[BasicSprite]] = []
        self._all_obstacles = Chain(self._walls, self._platforms)

        _add_to_list(self._ladders, ladders)
        _add_to_list(self._platforms, platforms)
        _add_to_list(self._walls, walls)

        self.player_sprite: Sprite = player_sprite
        """ The sprite controlled by the player.

        .. important:: This **must** be a :py:class:`.Sprite` or a
                       subclass of it!

        You can't use :py:class:`.BasicSprite` since it lacks
        required :py:attr:`~.Sprite.change_y` property.
        """
        self.gravity_constant: float = gravity_constant
        """The player's default downward acceleration.

        The engine's :py:meth:`update` method subtracts this value
        from the :py:attr:`player_sprite`'s :py:attr:`~.Sprite.change_y`
        when the player is not touching a sprite in :py:attr:`ladders` or
        :py:attr:`walls`.

        You can change the value of gravity after engine creation through
        this attribute. In addition to responding to GUI events, you can
        also change gravity in response to game events such as touching
        power-ups.

        Values for ``gravity_constant`` work as follows:

        .. list-table::
           :header-rows: 1

           * - ``gravity_constant``
             - Effect
           * - Greater than zero
             - Gravity points downward as expected
           * - Less than zero
             - Player falls upward (Consider adding a ceiling)
           * - Zero
             - No gravity

        To learn more, please see the following parts of the
        :ref:`platformer_tutorial`:

        * :ref:`platformer_part_five`
        * :ref:`platformer_part_twelve`
        """
        self.jumps_since_ground: int = 0
        """How many times the player has jumped since touching a
        sprite in :py:attr:`walls`.

        This is used throughout the engine's logic, including
        the :py:meth:`jump` and :py:meth:`can_jump` methods.
        """
        self.allowed_jumps: int = 1
        """Total number of jumps the player should be capable of.

        This includes the first jump. To enable multi-jump, see
        :py:meth:`enable_multi_jump` instead.
        """

        self.allow_multi_jump: bool = False
        """Whether multi-jump is enabled.

        For ease of use in simple games, you may want to use the
        following methods instead of setting this directly:

        * :py:meth:`enable_multi_jump`
        * :py:meth:`disable_multi_jump`
        """

    # The property object for ladders. This allows us setter/getter/deleter
    # capabilities in safe manner
    # TODO: figure out what do do with 15_ladders_moving_platforms.py
    # It's no longer used by any example or tutorial file
    @property
    def ladders(self) -> list[SpriteSequence[BasicSprite]]:
        """Ladders turn off gravity while touched by the player.

        This means that whenever the :py:attr:`player_sprite` collides
        with any any :py:class:`.Sprite` or :py:class:`.BasicSprite` in this
        list, the following are true:

        * The :py:attr:`gravity_constant` is not subtracted from
          :py:attr:`player_sprite`\'s :py:attr:`~.Sprite.change_y`
          during :py:meth:`update` calls
        * The player may otherwise move as freely as you allow

        """
        return self._ladders

    @ladders.setter
    def ladders(
        self,
        ladders: SpriteSequence[BasicSprite] | Iterable[SpriteSequence[BasicSprite]] | None = None,
    ) -> None:
        if ladders:
            _add_to_list(self._ladders, ladders)
        else:
            self._ladders.clear()

    @ladders.deleter
    def ladders(self) -> None:
        self._ladders.clear()

    @property
    def platforms(self) -> list[SpriteSequence[Sprite]]:
        """:py:class:`~arcade.sprite_list.sprite_list.SpriteList` instances containing platforms.

        .. important:: For best performance, put non-moving terrain in
                       :py:attr:`.walls` instead.

        Platforms are intended to support automatic movement by setting
        the appropriate attributes.

        You can enable automatic motion by setting one of the following
        attribute pairs on a :py:class:`~arcade.sprite.sprite.Sprite`:

        .. list-table::
           :header-rows: 1

           * - Movement Axis
             - :py:class:`~arcade.sprite.sprite.Sprite` Attributes to Set
           * - X (side to side)
             - * :py:attr:`~arcade.sprite.sprite.Sprite.change_x`
               * :py:attr:`~arcade.sprite.sprite.Sprite.boundary_left`
               * :py:attr:`~arcade.sprite.sprite.Sprite.boundary_right`
           * - Y (up and down)
             - * :py:attr:`~arcade.sprite.sprite.Sprite.change_y`
               * :py:attr:`~arcade.sprite.sprite.Sprite.boundary_bottom`
               * :py:attr:`~arcade.sprite.sprite.Sprite.boundary_top`

        For a working example, please see :ref:`sprite_moving_platforms`.
        """
        return self._platforms

    @platforms.setter
    def platforms(
        self, platforms: SpriteSequence[Sprite] | Iterable[SpriteSequence[Sprite]] | None = None
    ) -> None:
        if platforms:
            _add_to_list(self._platforms, platforms)
        else:
            self._platforms.clear()

    @platforms.deleter
    def platforms(self) -> None:
        self._platforms.clear()

    @property
    def walls(self) -> list[SpriteSequence[BasicSprite]]:
        """Exposes the :py:class:`SpriteList` instances use as terrain.

        .. important:: For best performance, only add non-moving sprites!

        The walls lists make a tradeoff through **spatial hashing**:

        * Collision checking against sprites in the list becomes
          very fast
        * Moving sprites or adding new ones becomes very slow

        This is worth the tradeoff for non-moving terrain, but it means
        you have to be careful. If you move too many sprites in the walls
        lists every frame, your game may slow down. For moving sprites
        the player can stand and jump on, see the :py:attr:`platforms`
        feature.

        To learn more about spatial hashing, please see the following:

        * :ref:`collision_detection_performance`
        * :py:class:`~arcade.sprite_list.spatial_hash.SpatialHash`
        """
        return self._walls

    @walls.setter
    def walls(
        self,
        walls: SpriteSequence[BasicSprite] | Iterable[SpriteSequence[BasicSprite]] | None = None,
    ) -> None:
        if walls:
            _add_to_list(self._walls, walls)
        else:
            self._walls.clear()

    @walls.deleter
    def walls(self) -> None:
        self._walls.clear()

    def is_on_ladder(self) -> bool:
        """Check if the :py:attr:`player_sprite` touches any :py:attr:`ladders`.

        .. warning:: This runs collisions **every** time it is called!

        Returns:
            ``True`` if the :py:attr:`player_sprite` touches any
            :py:attr:`ladders`.
        """
        if self.ladders:
            hit_list = check_for_collision_with_lists(self.player_sprite, self.ladders)
            if len(hit_list) > 0:
                return True
        return False

    def can_jump(self, y_distance: float = 5) -> bool:
        """Update jump state and return ``True`` if the player can jump.

        .. warning:: This runs collisions **every** time it is called!

        If you are thinking of calling this repeatedly, first double-check
        whether you can store the returned value to a local variable instead.

        The player can jump when at least one of the following are true:
        after updating state:

        .. list-table::
           :header-rows: 0

           * - The player is "touching" the ground
             - :py:attr:`player_sprite`\'s :py:attr:`.BasicSprite.center_y`
               is within ``y_distance`` of any sprite in :py:attr:`walls`
               or :py:attr:`platforms`

           * - The player can air-jump
             - :py:attr:`allow_multi_jump` is ``True`` and the player
               hasn't jumped more than :py:attr:`allowed_jumps` times

        Args:
            y_distance:
                The distance to temporarily move the :py:attr:`player_sprite`
                downward before checking for a collision with either
                :py:attr:`walls` or :py:attr:`platforms`.

        Returns:
             ``True`` if the player can jump.

        """

        # Temporarily move the player down to collide floor-like sprites
        self.player_sprite.center_y -= y_distance
        hit_list = check_for_collision_with_lists(self.player_sprite, self._all_obstacles)
        self.player_sprite.center_y += y_distance

        # Reset the number jumps if the player touched a floor-like sprite
        if len(hit_list) > 0:
            self.jumps_since_ground = 0

        if (
            len(hit_list) > 0
            or self.allow_multi_jump
            and self.jumps_since_ground < self.allowed_jumps
        ):
            return True
        else:
            return False

    def enable_multi_jump(self, allowed_jumps: int) -> None:
        """Enable multi-jump.

        The ``allowed_jumps`` argument is the total number of jumps
        the player should be able to make, including the first from
        solid ground in :py:attr:`walls` or any :py:attr:`platforms`.
        It will be stored as :py:attr:`allowed_jumps`.

        .. important:: If you override :py:meth:`jump`, be sure to call
                       :py:meth:`increment_jump_counter` inside it!

                       Otherwise, the player may be able to jump forever.

        Args:
            allowed_jumps:
                Total number of jumps the player should be capable of,
                including the first.

        """
        self.allowed_jumps = allowed_jumps
        self.allow_multi_jump = True

    def disable_multi_jump(self) -> None:
        """Disable multi-jump.

        Calling this function removes the requirement for :py:meth:`jump`
        to call :py:meth:`increment_jump_counter` with each jump to
        prevent infinite jumping.
        """
        self.allow_multi_jump = False
        self.allowed_jumps = 1
        self.jumps_since_ground = 0

    def jump(self, velocity: float) -> None:
        """Jump with an initial upward velocity.

        This works as follows:

        #. Set the :py:attr:`player_sprite`\'s :py:attr:`~.Sprite.change_y`
           to the passed ``velocity``
        #. Call :py:meth:`increment_jump_counter`

        Args:
            velocity:
                A positive value to set the player's y velocity to.
        """
        self.player_sprite.change_y = velocity
        self.increment_jump_counter()

    def increment_jump_counter(self) -> None:
        """Update jump tracking if multi-jump is enabled.

        If :py:attr:`allow_multi_jumps` is ``True``, calling this adds
        ``1`` to :py:attr:`jumps_since_ground`. Otherwise, it does
        nothing.
        """
        if self.allow_multi_jump:
            self.jumps_since_ground += 1

    def update(self) -> list[BasicSprite]:
        """Move the player and platforms, then return colliding sprites.

        The returned sprites will in a :py:class:`list` of individual
        sprites taken from all :py:class:`arcade.SpriteList` instances
        in the following:

        * :attr:`~platforms`
        * :attr:`~walls`

        The :py:attr:`~ladders` are not included.

        Returns:
            A list of all sprites the player collided with. If there
            were no collisions, the list will be empty.
        """
        # start_time = time.time()
        # print(f"Spot A ({self.player_sprite.center_x}, {self.player_sprite.center_y})")

        # --- Add gravity if we aren't on a ladder
        if not self.is_on_ladder():
            self.player_sprite.change_y -= self.gravity_constant

            # print(f"Spot F ({self.player_sprite.center_x}, {self.player_sprite.center_y})")

        # print(f"Spot B ({self.player_sprite.center_x}, {self.player_sprite.center_y})")

        for platform_list in self.platforms:
            for platform in platform_list:
                if platform.change_x != 0 or platform.change_y != 0:
                    # Check x boundaries and move the platform in x direction
                    if (
                        platform.boundary_left is not None
                        and platform.left <= platform.boundary_left
                    ):
                        platform.left = platform.boundary_left
                        if platform.change_x < 0:
                            platform.change_x *= -1

                    if (
                        platform.boundary_right is not None
                        and platform.right >= platform.boundary_right
                    ):
                        platform.right = platform.boundary_right
                        if platform.change_x > 0:
                            platform.change_x *= -1

                    platform.center_x += platform.change_x

                    # Check y boundaries and move the platform in y direction
                    if platform.boundary_top is not None and platform.top >= platform.boundary_top:
                        platform.top = platform.boundary_top
                        if platform.change_y > 0:
                            platform.change_y *= -1

                    if (
                        platform.boundary_bottom is not None
                        and platform.bottom <= platform.boundary_bottom
                    ):
                        platform.bottom = platform.boundary_bottom
                        if platform.change_y < 0:
                            platform.change_y *= -1

                    platform.center_y += platform.change_y

        complete_hit_list = _move_sprite(self.player_sprite, self._all_obstacles, ramp_up=True)

        # print(f"Spot Z ({self.player_sprite.center_x}, {self.player_sprite.center_y})")
        # Return list of encountered sprites
        # end_time = time.time()
        # print(f"Update - {end_time - start_time:7.4f}\n")

        return complete_hit_list



"""
Platformer Game

python -m arcade.examples.platform_tutorial.20_views
"""
import math

import arcade

# Constants
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Platformer"

# Constants used to scale our sprites from their original size
TILE_SCALING = 0.5
COIN_SCALING = 0.5

# Movement speed of player, in pixels per frame
PLAYER_MOVEMENT_SPEED = 5
GRAVITY = 1
PLAYER_JUMP_SPEED = 20

# Constants used to track the direction a character is facing
RIGHT_FACING = 0
LEFT_FACING = 1

class Character(arcade.Sprite):
    def __init__(self, name_folder, name_file):
        super().__init__()

        self.facing_direction = RIGHT_FACING

        self.cur_texture = 0

        main_path = f":resources:images/animated_characters/{name_folder}/{name_file}"
        # Load textures for idle, jump, and fall states
        idle_texture = arcade.load_texture(f"{main_path}_idle.png")
        jump_texture = arcade.load_texture(f"{main_path}_jump.png")
        fall_texture = arcade.load_texture(f"{main_path}_fall.png")
        # Make pairs with left and right facing textures
        self.idle_texture_pair = idle_texture, idle_texture.flip_left_right()
        self.jump_texture_pair = jump_texture, jump_texture.flip_left_right()
        self.fall_texture_pair = fall_texture, fall_texture.flip_left_right()
        # Load textures for walking with left and right facing textures
        self.walk_textures = []
        for i in range(8):
            texture = arcade.load_texture(f"{main_path}_walk{i}.png")
            self.walk_textures.append((texture, texture.flip_left_right()))

        self.climbing_textures = (
            arcade.load_texture(f"{main_path}_climb0.png"),
            arcade.load_texture(f"{main_path}_climb1.png")
        )

        # This variable will change dynamically and will represent the currently
        # active texture.
        self.texture = self.idle_texture_pair[0]


class PlayerCharacter(Character):
    def __init__(self):
        super().__init__("female_adventurer", "femaleAdventurer")

        # Track extra state related to the player. We will use these for change
        # textures in animations
        self.climbing = False
        self.should_update_walk = 0

    def update_animation(self, delta_time):

        # Figure out the direction the character is facing based on the movement
        # and previous direction.
        if self.change_x < 0 and self.facing_direction == RIGHT_FACING:
            self.facing_direction = LEFT_FACING
        elif self.change_x > 0 and self.facing_direction == LEFT_FACING:
            self.facing_direction = RIGHT_FACING

        # Handle animations for climbing on ladders. We use the absolute value
        # of change_y here because we don't care if the character is moving up
        # or down, the animation stays the same.
        if self.climbing and abs(self.change_y) > 1:
            self.cur_texture += 1
            if self.cur_texture > 7:
                self.cur_texture = 0
        if self.climbing:
            self.texture = self.climbing_textures[self.cur_texture // 4]
            return

        # Handling jumping animations
        if self.change_y > 0 and not self.climbing:
            self.texture = self.jump_texture_pair[self.facing_direction]
            return
        elif self.change_y < 0 and not self.climbing:
            self.texture = self.fall_texture_pair[self.facing_direction]
            return

        # Handle idle animations
        if self.change_x == 0:
            self.texture = self.idle_texture_pair[self.facing_direction]
            return

        # Handle walking
        if self.should_update_walk == 3:
            self.cur_texture += 1
            if self.cur_texture > 7:
                self.cur_texture = 0
            self.texture = self.walk_textures[self.cur_texture][self.facing_direction]
            self.should_update_walk = 0
            return

        self.should_update_walk += 1


class Enemy(Character):
    def __init__(self, name_folder, name_file):
        super().__init__(name_folder, name_file)

        self.should_update_walk = 0

    def update_animation(self, delta_time):
        # Figure out the direction the character is facing based on the
        # movement and previous direction.
        if self.change_x < 0 and self.facing_direction == RIGHT_FACING:
            self.facing_direction = LEFT_FACING
        elif self.change_x > 0 and self.facing_direction == LEFT_FACING:
            self.facing_direction = RIGHT_FACING

        # Handle idle animations
        if self.change_x == 0:
            self.texture = self.idle_texture_pair[self.facing_direction]
            return

        # Handle walking
        if self.should_update_walk == 3:
            self.cur_texture += 1
            if self.cur_texture > 7:
                self.cur_texture = 0
            self.texture = self.walk_textures[self.cur_texture][self.facing_direction]
            self.should_update_walk = 0
            return

        self.should_update_walk += 1


class RobotEnemy(Enemy):
    def __init__(self):
        super().__init__("robot", "robot")
        self.health = 100


class ZombieEnemy(Enemy):
    def __init__(self):
        super().__init__("zombie", "zombie")
        self.health = 50


class MainMenu(arcade.View):
    def on_show_view(self):
        self.window.background_color = arcade.color.WHITE

    def on_draw(self):
        self.clear()
        arcade.draw_text(
            "Main Menu - Click To Play",
            WINDOW_WIDTH // 2,
            WINDOW_HEIGHT // 2,
            arcade.color.BLACK,
            font_size=30,
            anchor_x="center"
        )

    def on_mouse_press(self, _x, _y, _button, _modifiers):
        game_view = GameView()
        self.window.show_view(game_view)


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):

        # Call the parent class and set up the window
        super().__init__()

        # Track the current state of our input
        self.left_pressed = False
        self.right_pressed = False
        self.up_pressed = False
        self.down_pressed = False
        self.shoot_pressed = False

        # Variable to hold our texture for our player
        self.player_texture = None

        # Separate variable that holds the player sprite
        self.player_sprite = None

        # Variable to hold our Tiled Map
        self.tile_map = None

        # Replacing all of our SpriteLists with a Scene variable
        self.scene = None

        # A variable to store our camera object
        self.camera = None

        # A variable to store our gui camera object
        self.gui_camera = None

        # This variable will store our score as an integer.
        self.score = 0

        # This variable will store the text for score that we will draw to the screen.
        self.score_text = None

        # Where is the right edge of the map?
        self.end_of_map = 0

        # Should we reset the score?
        self.reset_score = True

        # Shooting mechanics
        self.can_shoot = False
        self.shoot_timer = 0

        # Load sounds
        self.collect_coin_sound = arcade.load_sound(":resources:sounds/coin1.wav")
        self.jump_sound = arcade.load_sound(":resources:sounds/jump1.wav")
        self.gameover_sound = arcade.load_sound(":resources:sounds/gameover1.wav")
        self.shoot_sound = arcade.load_sound(":resources:sounds/hurt5.wav")
        self.hit_sound = arcade.load_sound(":resources:sounds/hit5.wav")

    def setup(self):
        """Set up the game here. Call this function to restart the game."""
        layer_options = {
            "Platforms": {
                "use_spatial_hash": True
            },
            "Moving Platforms": {
                "use_spatial_hash": False
            },
            "Ladders": {
                "use_spatial_hash": True
            }
        }

        # Load our TileMap
        self.tile_map = arcade.load_tilemap(
            ":resources:tiled_maps/map_with_ladders.json",
            scaling=TILE_SCALING,
            layer_options=layer_options,
        )

        # Create our Scene Based on the TileMap
        self.scene = arcade.Scene.from_tilemap(self.tile_map)

        self.player_sprite = PlayerCharacter()
        self.player_sprite.center_x = 128
        self.player_sprite.center_y = 128
        self.scene.add_sprite("Player", self.player_sprite)

        # -- Enemies
        enemies_layer = self.tile_map.object_lists["Enemies"]

        for enemy_marker in enemies_layer:
            coordinates = self.tile_map.get_cartesian(
                enemy_marker.shape[0], enemy_marker.shape[1]
            )
            enemy_type = enemy_marker.properties["type"]
            if enemy_type == "robot":
                enemy = RobotEnemy()
            elif enemy_type == "zombie":
                enemy = ZombieEnemy()
            enemy.center_x = math.floor(
                coordinates[0] * TILE_SCALING * self.tile_map.tile_width
            )
            enemy.center_y = math.floor(
                (coordinates[1] + 1) * (self.tile_map.tile_height * TILE_SCALING)
            )
            if "boundary_left" in enemy_marker.properties:
                enemy.boundary_left = enemy_marker.properties["boundary_left"]
            if "boundary_right" in enemy_marker.properties:
                enemy.boundary_right = enemy_marker.properties["boundary_right"]
            if "change_x" in enemy_marker.properties:
                enemy.change_x = enemy_marker.properties["change_x"]

            self.scene.add_sprite("Enemies", enemy)

        # Create a Platformer Physics Engine, this will handle moving our
        # player as well as collisions between the player sprite and
        # whatever SpriteList we specify for the walls.
        # It is important to supply static to the walls parameter. There is a
        # platforms parameter that is intended for moving platforms.
        # If a platform is supposed to move, and is added to the walls list,
        # it will not be moved.
        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite,
            walls=self.scene["Platforms"],
            gravity_constant=GRAVITY,
            platforms=self.scene["Moving Platforms"],
            ladders=self.scene["Ladders"]
        )

        # Initialize our camera, setting a viewport the size of our window.
        self.camera = arcade.Camera2D()

        # Initialize our gui camera, initial settings are the same as our world camera.
        self.gui_camera = arcade.Camera2D()

        # Reset the score if we should
        if self.reset_score:
            self.score = 0
        self.reset_score = True

        # Shooting mechanics
        self.can_shoot = False
        self.shoot_timer = 0

        # Initialize our arcade.Text object for score
        self.score_text = arcade.Text(f"Score: {self.score}", x=0, y=5)

        self.background_color = arcade.csscolor.CORNFLOWER_BLUE

        # Calculate the right edge of the map in pixels
        self.end_of_map = (self.tile_map.width * self.tile_map.tile_width)
        self.end_of_map *= self.tile_map.scaling

        # Add an empty bullet SpriteList to our scene
        self.scene.add_sprite_list("Bullets")

        self.window.background_color = self.tile_map.background_color

    def on_show_view(self):
        self.setup()

    def on_draw(self):
        """Render the screen."""

        # Clear the screen to the background color
        self.clear()

        # Activate our camera before drawing
        self.camera.use()

        # Draw our Scene
        self.scene.draw()

        # Activate our GUI camera
        self.gui_camera.use()

        # Draw our Score
        self.score_text.draw()

    def on_update(self, delta_time):
        """Movement and Game Logic"""

        # Move the player using our physics engine
        self.physics_engine.update()

        # Update our characters animation state
        if self.physics_engine.is_on_ladder():
            self.player_sprite.climbing = True
        else:
            self.player_sprite.climbing = False

        if self.can_shoot:
            if self.shoot_pressed:
                arcade.play_sound(self.shoot_sound)
                bullet = arcade.Sprite(
                    ":resources:images/space_shooter/laserBlue01.png",
                    scaling=0.8,
                )
                if self.player_sprite.facing_direction == RIGHT_FACING:
                    bullet.change_x = 12
                else:
                    bullet.change_x = -12

                bullet.center_x = self.player_sprite.center_x
                bullet.center_y = self.player_sprite.center_y

                self.scene.add_sprite("Bullets", bullet)
                self.can_shoot = False
        else:
            self.shoot_timer += 1
            if self.shoot_timer == 15:
                self.can_shoot = True
                self.shoot_timer = 0


        # Actually trigger animation updates. We've added the Background and Coins layer
        # here as well. Our Tiled map has some animated tiles built-in, check out the flags
        # and torches on the map.
        self.scene.update_animation(
            delta_time,
            [
                "Coins",
                "Background",
                "Player",
                "Enemies"
            ]
        )

        self.scene.update(delta_time, ["Enemies", "Bullets"])

        # Keep enemies walking within their boundaries configured in Tiled
        for enemy in self.scene["Enemies"]:
            if enemy.right > enemy.boundary_right and enemy.change_x > 0:
                enemy.change_x *= -1
            elif enemy.left < enemy.boundary_left and enemy.change_x < 0:
                enemy.change_x *= -1

        for bullet in self.scene["Bullets"]:
            hit_list = arcade.check_for_collision_with_lists(
                bullet,
                [
                    self.scene["Enemies"],
                    self.scene["Platforms"],
                    self.scene["Moving Platforms"]
                ]
            )

            if hit_list:
                bullet.remove_from_sprite_lists()

                for collision in hit_list:
                    if self.scene["Enemies"] in collision.sprite_lists:
                        collision.health -= 25

                        if collision.health <= 0:
                            collision.remove_from_sprite_lists()
                            self.score += 150

                        arcade.play_sound(self.hit_sound)

                return

            # Remove bullet if it leaves the map area.
            # Bullets only travel horizontally, so we only need to check left and right.
            if (bullet.right < 0) or (bullet.left > self.end_of_map):
                bullet.remove_from_sprite_lists()

        # See if we hit any coins
        player_collision_list = arcade.check_for_collision_with_lists(
            self.player_sprite,
            [
                self.scene["Coins"],
                self.scene["Enemies"]
            ]
        )

        for collision in player_collision_list:
            if self.scene["Enemies"] in collision.sprite_lists:
                arcade.play_sound(self.gameover_sound)
                game_over = GameOverView()
                self.window.show_view(game_over)
                return
            else:
                # Our collision is a coin, remove it
                collision.remove_from_sprite_lists()
                arcade.play_sound(self.collect_coin_sound)
                self.score += 75
                self.score_text.text = f"Score: {self.score}"

        # Center our camera on the player
        self.camera.position = self.player_sprite.position

    def process_keychange(self):
        # First handle the case where we have moved up. This needs to be handled
        # differently to move the player upwards if they are on a ladder, or
        # perform a jump if they are not on a ladder. This code might look
        # different if we had a separate button for jumping, we would only need
        # to handle moving upwards if we were on a ladder for the up key then.
        # Here we also handle the case where we have moved down while on a ladder.
        if self.up_pressed and not self.down_pressed:
            if self.physics_engine.is_on_ladder():
                self.player_sprite.change_y = PLAYER_MOVEMENT_SPEED
            elif self.physics_engine.can_jump(y_distance=10):
                self.player_sprite.change_y = PLAYER_JUMP_SPEED
                arcade.play_sound(self.jump_sound)
        elif self.down_pressed and not self.up_pressed:
            if self.physics_engine.is_on_ladder():
                self.player_sprite.change_y = -PLAYER_MOVEMENT_SPEED

        # Now we need a special handling of our vertical movement while we are 
        # on a ladder, but have no input specified. When we jump, the physics
        # engine takes care of resetting our vertical movement to zero once we've
        # hit the ground. However for ladders, we need to ensure that we set the
        # vertical movement back to zero if the user does not give input, otherwise
        # once a user starts climbing a ladder, they will move upwards automatically
        # until they reach the end of the ladder. You can try commenting out this
        # block to see what that effect looks like.
        if self.physics_engine.is_on_ladder():
            if not self.up_pressed and not self.down_pressed:
                self.player_sprite.change_y = 0
            elif self.up_pressed and self.down_pressed:
                self.player_sprite.change_y = 0

        # Now we just handle our horizontal movement, very similar to how we
        # did before, but now just combined in our new function.
        if self.right_pressed and not self.left_pressed:
            self.player_sprite.change_x = PLAYER_MOVEMENT_SPEED
        elif self.left_pressed and not self.right_pressed:
            self.player_sprite.change_x = -PLAYER_MOVEMENT_SPEED
        else:
            self.player_sprite.change_x = 0

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed."""

        if key == arcade.key.ESCAPE:
            self.setup()

        if key == arcade.key.UP or key == arcade.key.W:
            self.up_pressed = True
        elif key == arcade.key.DOWN or key == arcade.key.S:
            self.down_pressed = True
        elif key == arcade.key.LEFT or key == arcade.key.A:
            self.left_pressed = True
        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.right_pressed = True

        if key == arcade.key.Q:
            self.shoot_pressed = True

        self.process_keychange()

    def on_key_release(self, key, modifiers):
        """Called whenever a key is released."""

        if key == arcade.key.LEFT or key == arcade.key.A:
            self.left_pressed = False
        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.right_pressed = False
        elif key == arcade.key.UP or key == arcade.key.W:
            self.up_pressed = False
        elif key == arcade.key.DOWN or key == arcade.key.S:
            self.down_pressed = False

        if key == arcade.key.Q:
            self.shoot_pressed = False

        self.process_keychange()


class GameOverView(arcade.View):
    def on_show_view(self):
        self.window.background_color = arcade.color.BLACK

    def on_draw(self):
        self.clear()
        arcade.draw_text(
            "Game Over - Click to Restart",
            WINDOW_WIDTH // 2,
            WINDOW_HEIGHT // 2,
            arcade.color.WHITE,
            30,
            anchor_x="center"
        )

    def on_mouse_press(self, _x, _y, _button, _modifiers):
        game_view = GameView()
        self.window.show_view(game_view)

def main():
    """Main function"""
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    menu_view = MainMenu()
    window.show_view(menu_view)
    arcade.run()


if __name__ == "__main__":
    main()



""":py:class:`typing.TypedDict` entries.

Placing them here prevents circular import issues.
"""

from typing_extensions import TypedDict


class RawBindBase(TypedDict):
    """General base for raw axis or action binds.

    Anything matching this can be passed to
    :py:func:`~arcade.future.input.inputs.parse_mapping_input_enum` to
    extract a corresponding :py:class:`~arcade.future.input.inputs.InputEnum`
    value.

    For specific raw types, see:

    * :py:class:`RawActionMapping`
    * :py:class:`RawAxisMapping`
    """

    input_type: int
    input: str | int


class RawActionMapping(RawBindBase):
    mod_shift: bool
    mod_ctrl: bool
    mod_alt: bool


class RawAxisMapping(RawBindBase):
    scale: float


class RawAction(TypedDict):
    """Annotates the raw form for :py:class:`ActionMapping`."""

    name: str
    mappings: list[RawActionMapping]


class RawAxis(TypedDict):
    """Annotates the raw form for :py:class:`AxisMapping`."""

    name: str
    mappings: list[RawAxisMapping]



from collections.abc import Callable
from typing import Generic, TypeVar, cast

import arcade
from arcade.color import BLACK, RED, WHITE
from arcade.gui import UIEvent, UIInputText, UILabel, UITextInputEvent
from arcade.types import Color, RGBOrA255
from arcade.utils import type_name

__all__ = ("UITypedTextInput",)


T = TypeVar("T")


# This is almost certainly doing at least one thing wrong
class UITypedTextInput(UIInputText, Generic[T]):
    """A text box which auto-converts to and from a :py:class:`type`.

    The simplest usage is passing a :py:func:`type` which supports
    :py:func:`repr` and allows a single :py:class:`str` as an argument:

    .. code-block:: python

       self.float_input = UITypedTextInput(float, text="0.0")

    In the example above, setting :py:attr:`self.float_input.text` to
    ``"string"`` will:

    #. Set both the text and the caret to the ``error_color`` passed at
       creation
    #. Re-raise the :py:class:`ValueError` from ``float("string")``

    To stop error propagation, pass
    You can customize your conversion to and from strings by overriding
    the following arguments with custom :py:class:`callable` objects:

    .. list-table::
       :header-rows: 1

       * - Argument
         - Default

       * - ``to_str``
         - :py:func:`repr`

       * - ``from_str``
         - the ``parsed_type``

    .. important:: This class is meant to handle simple types in simple
                   dev and test tools.

                   As a general rule, if you need to highlight a specific
                   syntax error, this class is not the right tool.

    Args:
        parsed_type:
            The :py:class:`type` to require. This is not meant to
            be changed after creation.
        from_str:
            A type or other :py:func:`callable` which converts a
            :py:class:`str` to an instance of :py:class:`parsed_type`.
            It may raise exceptions and perform cleaning of text.
        to_str:
            A :py:func:`callable` which converts ``parsed_type``
            instances to :py:class:`str`.
        x: an X position (see :py:class:`.UIInputText`).
        y: an X position (see :py:class:`.UIInputText`).
        width: an X axis width (see :py:class:`.UIInputText`).
        height: a Y axis height (see :py:class:`.UIInputText`).
        text: The initial text to display.
        font_name: (see :py:class:`.UIInputText`).
        text_color: The color to use for non-error text.
        error_color: The color to use when ``to_str`` or ``from_str``
            raised an exception.
        multiline: See :py:class:`.UIInputText`.
        size_hint:  See :py:class:`.UIInputText`.
        size_hint_min:  See :py:class:`.UIInputText`.
        size_hint_max: See :py:class:`.UIInputText`.
    """

    def __init__(
        self,
        parsed_type: type[T],
        *,
        to_str: Callable[[T], str] = repr,
        from_str: Callable[[str], T] | None = None,
        emit_parse_exceptions: bool = True,
        x: float = 0,
        y: float = 0,
        width: float = 100,
        height: float = 24,
        text: str = "",
        font_name=("Arial",),
        font_size: float = 12,
        text_color: RGBOrA255 = BLACK,
        error_color: RGBOrA255 = RED,
        multiline=False,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        **kwargs,
    ):
        if not isinstance(type, type):
            raise TypeError(f"Expected a type, but got {parsed_type}")
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            text=text,
            font_name=font_name,
            font_size=font_size,
            text_color=text_color,
            multiline=multiline,
            caret_color=text_color,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )
        self.emit_parse_exceptions = emit_parse_exceptions
        self._error_color = error_color
        self._valid_color = text_color
        self._parsed_type: type[T] = parsed_type
        self._to_str = to_str
        self._from_str: Callable[[str], T] = cast(Callable[[str], T], from_str or parsed_type)
        self._parsed_value: T = self._from_str(self.text)

    def _set_current_color(self, new_color: RGBOrA255) -> None:
        validated = Color.from_iterable(new_color)
        if self._text_color == validated:
            return

        self._text_color = validated
        self.caret.color = validated
        self.doc.set_style(0, len(self.text), dict(color=validated))
        self.trigger_full_render()

    def _checked_parse(self, text: str):
        try:
            self._parsed_value = self._from_str(text)
            self._set_current_color(self._valid_color)
        except Exception as e:
            # print(e)
            self._set_current_color(self._error_color)
            if self.emit_parse_exceptions:
                raise e

    def on_event(self, event: UIEvent) -> bool | None:
        # print(f"In {type_name(event)}")
        if isinstance(event, UITextInputEvent) and self._active:
            text = event.text.replace("\r", "").replace("\r", "")
            event.text = text

        handled = super().on_event(event)
        self._checked_parse(self.doc.text)
        return handled

    @property
    def parsed_type(self) -> type[T]:
        """Get the type this input field expects to parse.

        .. note:: This is not meant to be changed after creation.
        """
        return self._parsed_type

    @property
    def value(self) -> T:
        """The current instance of :py:attr:`parsed_type`.

        Setting this automatically updates the text of the widget.
        """
        return self._parsed_value

    @value.setter
    def value(self, new_value: T) -> None:
        must_be = self._parsed_type
        if not isinstance(new_value, must_be):
            raise TypeError(
                # We pass self here to support subclasses
                f"This {type_name(self)} was created to expect {type_name(must_be)}"
                f", but got {new_value!r} (a {type_name(new_value)})"
            )
        self.doc.text = self._to_str(new_value)
        self._set_current_color(self._valid_color)

    @property
    def text(self) -> str:
        """Get/set the text of the widget.

        In addition to basic behavior from :py:class:`UITextWidget`,
        this also performs validation. To silence error propagation
        from validation, set :py:attr:`emit_parse_exceptions` to
        ``False``.
        """
        return self.doc.text

    @text.setter
    def text(self, new_text: str) -> None:
        self.doc.text = new_text
        self._checked_parse(new_text)


if __name__ == "__main__":
    width, height = 400, 400
    center = width / 2
    from arcade.gui import NinePatchTexture, UIView

    class MyView(UIView):
        def __init__(self):
            super().__init__()
            self.ninepatch = NinePatchTexture(
                left=5,
                right=5,
                top=5,
                bottom=5,
                texture=arcade.load_texture(":resources:gui_basic_assets/window/panel_gray.png"),
            )
            self.instructions = UILabel(
                text="Valid float values -> black text\nNon-float values -> red text",
                x=center - 150,
                height=center + 80,
                text_color=WHITE,
                multiline=True,
                font_size=16,
                width=300,
                align="center",
            )
            self.ui.add(self.instructions)

            self.float_box = UITypedTextInput(
                float,
                text="0.0",
                x=center - 100,
                y=height / 2,
                width=200,
                font_size=16,
                height=30,
                emit_parse_exceptions=False,
            ).with_background(texture=self.ninepatch)
            self.ui.add(self.float_box)

        def on_show_view(self):
            super().on_show_view()

    window = arcade.Window(width, height, "Typed input text test")
    window.show_view(MyView())
    window.run()



"""
Various utility functions for the gl module.
"""

from array import array
from ctypes import c_byte
from typing import Any


def data_to_ctypes(data: Any) -> tuple[int, Any]:
    """
    Attempts to convert the data to ctypes if needed by using the buffer protocol.

    - bytes will be returned as is
    - Tuples will be converted to array
    - Other types will be converted to ctypes by using the buffer protocol
      by creating a memoryview and then a ctypes array of bytes.

    Args:
        data: The data to convert to ctypes.
    Returns:
        A tuple containing the size of the data in bytes
        and the data object optionally converted to ctypes.
    """
    if isinstance(data, bytes):
        return len(data), data
    else:
        if isinstance(data, tuple):
            data = array("f", data)
        try:
            m_view = memoryview(data)
            c_bytes = c_byte * m_view.nbytes
            return m_view.nbytes, c_bytes.from_buffer(m_view)
        except Exception as ex:
            raise TypeError(f"Failed to convert data to ctypes: {ex}")



"""
Platformer Game

python -m arcade.examples.platform_tutorial.11_scene
"""
import arcade

# Constants
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Platformer"

# Constants used to scale our sprites from their original size
TILE_SCALING = 0.5
COIN_SCALING = 0.5

# Movement speed of player, in pixels per frame
PLAYER_MOVEMENT_SPEED = 5
GRAVITY = 1
PLAYER_JUMP_SPEED = 20


class GameView(arcade.Window):
    """
    Main application class.
    """

    def __init__(self):

        # Call the parent class and set up the window
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

        # Variable to hold our texture for our player
        self.player_texture = None

        # Separate variable that holds the player sprite
        self.player_sprite = None

        # Replacing all of our SpriteLists with a Scene variable
        self.scene = None

        # A variable to store our camera object
        self.camera = None

        # A variable to store our gui camera object
        self.gui_camera = None

        # This variable will store our score as an integer.
        self.score = 0

        # This variable will store the text for score that we will draw to the screen.
        self.score_text = None

        # Load sounds
        self.collect_coin_sound = arcade.load_sound(":resources:sounds/coin1.wav")
        self.jump_sound = arcade.load_sound(":resources:sounds/jump1.wav")

    def setup(self):
        """Set up the game here. Call this function to restart the game."""
        self.scene = arcade.Scene()

        self.player_texture = arcade.load_texture(
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_idle.png"
        )

        self.player_sprite = arcade.Sprite(self.player_texture)
        self.player_sprite.center_x = 64
        self.player_sprite.center_y = 128
        self.scene.add_sprite("Player", self.player_sprite)

        self.scene.add_sprite_list("Walls", use_spatial_hash=True)
        self.scene.add_sprite_list("Coins", use_spatial_hash=True)

        # Create the ground
        # This shows using a loop to place multiple sprites horizontally
        for x in range(0, 1250, 64):
            wall = arcade.Sprite(":resources:images/tiles/grassMid.png", scale=TILE_SCALING)
            wall.center_x = x
            wall.center_y = 32
            self.scene.add_sprite("Walls", wall)

        # Put some crates on the ground
        # This shows using a coordinate list to place sprites
        coordinate_list = [[512, 96], [256, 96], [768, 96]]

        for coordinate in coordinate_list:
            # Add a crate on the ground
            wall = arcade.Sprite(
                ":resources:images/tiles/boxCrate_double.png", scale=TILE_SCALING
            )
            wall.position = coordinate
            self.scene.add_sprite("Walls", wall)

        # Add coins to the world
        for x in range(128, 1250, 256):
            coin = arcade.Sprite(":resources:images/items/coinGold.png", scale=COIN_SCALING)
            coin.center_x = x
            coin.center_y = 96
            self.scene.add_sprite("Coins", coin)

        # Create a Platformer Physics Engine, this will handle moving our
        # player as well as collisions between the player sprite and
        # whatever SpriteList we specify for the walls.
        # It is important to supply static to the walls parameter. There is a
        # platforms parameter that is intended for moving platforms.
        # If a platform is supposed to move, and is added to the walls list,
        # it will not be moved.
        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite, walls=self.scene["Walls"], gravity_constant=GRAVITY
        )

        # Initialize our camera, setting a viewport the size of our window.
        self.camera = arcade.Camera2D()

        # Initialize our gui camera, initial settings are the same as our world camera.
        self.gui_camera = arcade.Camera2D()

        # Reset our score to 0
        self.score = 0

        # Initialize our arcade.Text object for score
        self.score_text = arcade.Text(f"Score: {self.score}", x=0, y=5)

        self.background_color = arcade.csscolor.CORNFLOWER_BLUE

    def on_draw(self):
        """Render the screen."""

        # Clear the screen to the background color
        self.clear()

        # Activate our camera before drawing
        self.camera.use()

        # Draw our Scene
        self.scene.draw()

        # Activate our GUI camera
        self.gui_camera.use()

        # Draw our Score
        self.score_text.draw()

    def on_update(self, delta_time):
        """Movement and Game Logic"""

        # Move the player using our physics engine
        self.physics_engine.update()

        # See if we hit any coins
        coin_hit_list = arcade.check_for_collision_with_list(
            self.player_sprite, self.scene["Coins"]
        )

        # Loop through each coin we hit (if any) and remove it
        for coin in coin_hit_list:
            # Remove the coin
            coin.remove_from_sprite_lists()
            arcade.play_sound(self.collect_coin_sound)
            self.score += 75
            self.score_text.text = f"Score: {self.score}"

        # Center our camera on the player
        self.camera.position = self.player_sprite.position

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed."""

        if key == arcade.key.ESCAPE:
            self.setup()

        if key == arcade.key.UP or key == arcade.key.W:
            if self.physics_engine.can_jump():
                self.player_sprite.change_y = PLAYER_JUMP_SPEED
                arcade.play_sound(self.jump_sound)

        if key == arcade.key.LEFT or key == arcade.key.A:
            self.player_sprite.change_x = -PLAYER_MOVEMENT_SPEED
        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.player_sprite.change_x = PLAYER_MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """Called whenever a key is released."""

        if key == arcade.key.LEFT or key == arcade.key.A:
            self.player_sprite.change_x = 0
        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.player_sprite.change_x = 0


def main():
    """Main function"""
    window = GameView()
    window.setup()
    arcade.run()


if __name__ == "__main__":
    main()



"""Example of using experimental UIRestrictedInputText.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.exp_restricted_input
"""

import arcade
from arcade.gui import UIAnchorLayout, UIBoxLayout, UIView
from arcade.gui.experimental.restricted_input import UIIntInput


class MyView(UIView):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_BELIZE_HOLE

        root = self.ui.add(UIAnchorLayout())
        bars = root.add(UIBoxLayout(space_between=10))

        # UIWidget based progress bar
        self.input_field = UIIntInput(width=300, height=40, font_size=22)
        bars.add(self.input_field)


def main():
    window = arcade.Window(antialiasing=False)
    window.show_view(MyView())
    arcade.run()


if __name__ == "__main__":
    main()



from .particle import (
    Particle,
    EternalParticle,
    LifetimeParticle,
    FadeParticle,
)
from .emitter import (
    Emitter,
    EmitController,
    EmitBurst,
    EmitMaintainCount,
    EmitInterval,
    EmitterIntervalWithCount,
    EmitterIntervalWithTime,
)
from .emitter_simple import (
    make_burst_emitter,
    make_interval_emitter,
)

__all__ = [
    "Particle",
    "EternalParticle",
    "LifetimeParticle",
    "FadeParticle",
    "Emitter",
    "EmitController",
    "EmitBurst",
    "EmitMaintainCount",
    "EmitInterval",
    "EmitterIntervalWithCount",
    "EmitterIntervalWithTime",
    "make_burst_emitter",
    "make_interval_emitter",
]



"""
Sprites with Properties Example

Simple program to show how to store properties on sprites.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_properties
"""
import arcade

# --- Constants ---
SPRITE_SCALING_PLAYER = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprites with Properties Example"


INSTRUCTIONS1 = (
    "Touch a coin to set its intensity property to 'bright'."
    "Press 'R' to reset the sprites"
)
INSTRUCTIONS2 = "Touch the trigger at the bottom-right to destroy all 'bright' sprites."


class GameView(arcade.View):
    """ Our custom Window Class"""

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = None
        self.coin_list = None

        # Set up the player info
        self.player_sprite = None

        # Set up sprite that will serve as trigger
        self.trigger_sprite = None

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        self.background_color = arcade.color.AMAZON

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.coin_list = arcade.SpriteList()

        # Set up the player
        # Character image from kenney.nl
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=0.75,
        )
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 150
        self.player_list.append(self.player_sprite)

        # Create the sprites
        for x in range(180, 1100, 100):
            coin = arcade.Sprite(
                ":resources:images/items/coinGold.png",
                scale=0.3,
                center_x=x,
                center_y=400,
            )
            coin.intensity = 'dim'
            coin.alpha = 64
            self.coin_list.append(coin)

        # Create trigger
        self.trigger_sprite = arcade.Sprite(
            ":resources:images/pinball/bumper.png", scale=0.5,
            center_x=750, center_y=50,
        )

    def on_key_press(self, symbol, modifiers):
        if symbol == arcade.key.R:
            self.setup()

    def on_draw(self):
        """ Draw everything """
        self.clear()
        self.coin_list.draw()
        arcade.draw_sprite(self.trigger_sprite)
        self.player_list.draw()

        # Put the instructions on the screen.
        arcade.draw_text(INSTRUCTIONS1, 10, 90, arcade.color.WHITE, 14)
        arcade.draw_text(INSTRUCTIONS2, 10, 70, arcade.color.WHITE, 14)

        # Query the property on the coins and show results.
        coins_are_bright = [coin.intensity == 'bright' for coin in self.coin_list]
        output_any = f"Any sprites have intensity=bright? : {any(coins_are_bright)}"
        arcade.draw_text(output_any, 10, 40, arcade.color.WHITE, 14)
        output_all = f"All sprites have intensity=bright? : {all(coins_are_bright)}"
        arcade.draw_text(output_all, 10, 20, arcade.color.WHITE, 14)

    def on_mouse_motion(self, x, y, dx, dy):
        """ Handle Mouse Motion """

        # Move the center of the player sprite to match the mouse x, y
        self.player_sprite.center_x = x
        self.player_sprite.center_y = y

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.coin_list.update()

        # Generate a list of all sprites that collided with the player.
        coins_hit_list = arcade.check_for_collision_with_list(
            self.player_sprite, self.coin_list,
        )

        # Loop through each colliding sprite to set intensity=bright
        for coin in coins_hit_list:
            coin.intensity = 'bright'
            coin.alpha = 255

        hit_trigger = arcade.check_for_collision(self.player_sprite, self.trigger_sprite)
        if hit_trigger:
            intense_sprites = [
                sprite for sprite in self.coin_list if sprite.intensity == 'bright'
            ]
            for coin in intense_sprites:
                coin.remove_from_sprite_lists()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



import struct
from collections.abc import Iterable

from arcade import (
    get_window,
)
from arcade.geometry import (
    are_polygons_intersecting,
    is_point_in_polygon,
)
from arcade.math import get_distance
from arcade.sprite import BasicSprite, SpriteType
from arcade.types import Point
from arcade.types.rect import Rect

from .sprite_list import SpriteSequence


def get_distance_between_sprites(sprite1: SpriteType, sprite2: SpriteType) -> float:
    """
    Returns the distance between the center of two given sprites

    Args:
        sprite1: Sprite one
        sprite2: Sprite two
    """
    return get_distance(*sprite1._position, *sprite2._position)


def get_closest_sprite(
    sprite: BasicSprite, sprite_list: SpriteSequence[SpriteType]
) -> tuple[SpriteType, float] | None:
    """
    Given a Sprite and SpriteList, returns the closest sprite, and its distance.

    Args:
        sprite:
            Target sprite
        sprite_list:
            List to search for closest sprite.

    Returns:
        A tuple containing the closest sprite and the minimum distance.
        If the spritelist is empty we return ``None``.
    """
    if len(sprite_list) == 0:
        return None

    min_pos = 0
    min_distance = get_distance_between_sprites(sprite, sprite_list[min_pos])
    for i in range(1, len(sprite_list)):
        distance = get_distance_between_sprites(sprite, sprite_list[i])
        if distance < min_distance:
            min_pos = i
            min_distance = distance
    return sprite_list[min_pos], min_distance


def check_for_collision(sprite1: BasicSprite, sprite2: BasicSprite) -> bool:
    """
    Check for a collision between two sprites.

    Args:
        sprite1: First sprite
        sprite2: Second sprite

    Returns:
        ``True`` or ``False`` depending if the sprites intersect.
    """
    if __debug__:
        if not isinstance(sprite1, BasicSprite):
            raise TypeError("Parameter 1 is not an instance of a Sprite class.")
        if isinstance(sprite2, SpriteSequence):
            raise TypeError(
                "Parameter 2 is a instance of the SpriteList instead of a required "
                "Sprite. See if you meant to call check_for_collision_with_list instead "
                "of check_for_collision."
            )
        elif not isinstance(sprite2, BasicSprite):
            raise TypeError("Parameter 2 is not an instance of a Sprite class.")

    return _check_for_collision(sprite1, sprite2)


def _check_for_collision(sprite1: BasicSprite, sprite2: BasicSprite) -> bool:
    """
    Check for collision between two sprites.

    Args:
        sprite1: Sprite 1
        sprite2: Sprite 2
    Returns:
        ``True`` if sprites overlap.
    """

    # NOTE: for speed because attribute look ups are slow.
    sprite1_position = sprite1._position
    sprite1_width = sprite1._width
    sprite1_height = sprite1._height
    sprite2_position = sprite2._position
    sprite2_width = sprite2._width
    sprite2_height = sprite2._height

    radius_sum = (sprite1_width if sprite1_width > sprite1_height else sprite1_height) + (
        sprite2_width if sprite2_width > sprite2_height else sprite2_height
    )

    # Multiply by half of the theoretical max diagonal length for an estimation of distance
    radius_sum *= 0.71  # 1.42 / 2
    radius_sum_sq = radius_sum * radius_sum

    diff_x = sprite1_position[0] - sprite2_position[0]
    diff_x_sq = diff_x * diff_x
    if diff_x_sq > radius_sum_sq:
        return False

    diff_y = sprite1_position[1] - sprite2_position[1]
    diff_y_sq = diff_y * diff_y
    if diff_y_sq > radius_sum_sq:
        return False

    distance = diff_x_sq + diff_y_sq
    if distance > radius_sum_sq:
        return False

    return are_polygons_intersecting(
        sprite1.hit_box.get_adjusted_points(), sprite2.hit_box.get_adjusted_points()
    )


def _get_nearby_sprites(
    sprite: BasicSprite, sprite_list: SpriteSequence[SpriteType]
) -> list[SpriteType]:
    sprite_count = len(sprite_list)
    if sprite_count == 0:
        return []

    # Update the position and size to check
    ctx = get_window().ctx
    sprite_list._write_sprite_buffers_to_gpu()

    ctx.collision_detection_program["check_pos"] = sprite.position
    ctx.collision_detection_program["check_size"] = sprite.width, sprite.height

    # Ensure the result buffer can fit all the sprites (worst case)
    buffer = ctx.collision_buffer
    if buffer.size < sprite_count * 4:
        buffer.orphan(size=sprite_count * 4)

    # Run the transform shader emitting sprites close to the configured position and size.
    # This runs in a query so we can measure the number of sprites emitted.
    with ctx.collision_query:
        sprite_list.geometry.transform(  # type: ignore
            ctx.collision_detection_program,
            buffer,
            vertices=sprite_count,
        )

    # Store the number of sprites emitted
    emit_count = ctx.collision_query.primitives_generated
    # print(
    #     emit_count,
    #     ctx.collision_query.time_elapsed,
    #     ctx.collision_query.time_elapsed / 1_000_000_000,
    # )

    # If no sprites emitted we can just return an empty list
    if emit_count == 0:
        return []

    # # Debug block for transform data to keep around
    # print("emit_count", emit_count)
    # data = buffer.read(size=emit_count * 4)
    # print("bytes", data)
    # print("data", struct.unpack(f'{emit_count}i', data))

    # .. otherwise build and return a list of the sprites selected by the transform
    return [
        sprite_list[i] for i in struct.unpack(f"{emit_count}i", buffer.read(size=emit_count * 4))
    ]


def check_for_collision_with_list(
    sprite: BasicSprite,
    sprite_list: SpriteSequence[SpriteType],
    method: int = 0,
) -> list[SpriteType]:
    """
    Check for a collision between a sprite, and a list of sprites.

    Args:
        sprite:
            Sprite to check
        sprite_list:
            SpriteList to check against
        method:
            Collision check method. Defaults to 0.

            - 0: auto-select. (spatial if available, GPU if 1500+ sprites, else simple)
            - 1: Spatial Hashing if available,
            - 2: GPU based
            - 3: Simple check-everything.

            Note that while the GPU method is very fast when you cannot use spatial hashing,
            it's also very slow if you are calling this function many times per frame.
            What method is the most appropriate depends entirely on your use case.

    Returns:
        List of sprites colliding, or an empty list.
    """
    if __debug__:
        if not isinstance(sprite, BasicSprite):
            raise TypeError(
                f"Parameter 1 is not an instance of the Sprite class, "
                f"it is an instance of {type(sprite)}."
            )
        if not isinstance(sprite_list, SpriteSequence):
            raise TypeError(f"Parameter 2 is a {type(sprite_list)} instead of expected SpriteList.")

    sprites_to_check: Iterable[SpriteType]
    # Spatial
    if sprite_list.spatial_hash is not None and (method == 1 or method == 0):
        sprites_to_check = sprite_list.spatial_hash.get_sprites_near_sprite(sprite)
    elif method == 3 or (method == 0 and len(sprite_list) <= 1500):
        sprites_to_check = sprite_list
    else:
        # GPU transform
        sprites_to_check = _get_nearby_sprites(sprite, sprite_list)

    return [
        sprite2
        for sprite2 in sprites_to_check
        if sprite is not sprite2 and _check_for_collision(sprite, sprite2)
    ]

    # collision_list = []
    # for sprite2 in sprite_list_to_check:
    #     if sprite1 is not sprite2 and sprite2 not in collision_list:
    #         if _check_for_collision(sprite1, sprite2):
    #             collision_list.append(sprite2)


def check_for_collision_with_lists(
    sprite: BasicSprite,
    sprite_lists: Iterable[SpriteSequence[SpriteType]],
    method=1,
) -> list[SpriteType]:
    """
    Check for a collision between a Sprite, and a list of SpriteLists.

    Args:
        sprite:
            Sprite to check
        sprite_lists:
            SpriteLists to check against
        method:
            Collision check method. 1 is Spatial Hashing if available,
            2 is GPU based, 3 is slow CPU-bound check-everything. Defaults to 1.

    Returns:
        List of sprites colliding, or an empty list.
    """
    if __debug__:
        if not isinstance(sprite, BasicSprite):
            raise TypeError(
                f"Parameter 1 is not an instance of the BasicSprite class, "
                f"it is an instance of {type(sprite)}."
            )

    sprites: list[SpriteType] = []
    sprites_to_check: Iterable[SpriteType]

    for sprite_list in sprite_lists:
        if sprite_list.spatial_hash is not None and method == 1:
            sprites_to_check = sprite_list.spatial_hash.get_sprites_near_sprite(sprite)
        elif method == 3:
            sprites_to_check = sprite_list
        else:
            # GPU transform
            sprites_to_check = _get_nearby_sprites(sprite, sprite_list)

        for sprite2 in sprites_to_check:
            if sprite is not sprite2 and _check_for_collision(sprite, sprite2):
                sprites.append(sprite2)

    return sprites


def get_sprites_at_point(point: Point, sprite_list: SpriteSequence[SpriteType]) -> list[SpriteType]:
    """
    Get a list of sprites at a particular point. This function sees if any sprite overlaps
    the specified point. If a sprite has a different center_x/center_y but touches the point,
    this will return that sprite.

    Args:
        point: Point to check
        sprite_list: SpriteList to check against

    :returns: List of sprites colliding, or an empty list.
    """
    if __debug__:
        if not isinstance(sprite_list, SpriteSequence):
            raise TypeError(f"Parameter 2 is a {type(sprite_list)} instead of expected SpriteList.")

    sprites_to_check: Iterable[SpriteType]

    if sprite_list.spatial_hash is not None:
        sprites_to_check = sprite_list.spatial_hash.get_sprites_near_point(point)
    else:
        sprites_to_check = sprite_list

    return [
        s
        for s in sprites_to_check
        if is_point_in_polygon(point[0], point[1], s.hit_box.get_adjusted_points())
    ]


def get_sprites_at_exact_point(
    point: Point, sprite_list: SpriteSequence[SpriteType]
) -> list[SpriteType]:
    """
    Get a list of sprites whose center_x, center_y match the given point.
    This does NOT return sprites that overlap the point, the center has to be an exact match.

    Args:
        point: Point to check
        sprite_list: SpriteList to check against
    Returns:
        List of sprites colliding, or an empty list.
    """
    if __debug__:
        if not isinstance(sprite_list, SpriteSequence):
            raise TypeError(f"Parameter 2 is a {type(sprite_list)} instead of expected SpriteList.")

    sprites_to_check: Iterable[SpriteType]

    if sprite_list.spatial_hash is not None:
        sprites_to_check = sprite_list.spatial_hash.get_sprites_near_point(point)
        # checks_saved = len(sprite_list) - len(sprite_list_to_check)
        # print("Checks saved: ", checks_saved)
    else:
        sprites_to_check = sprite_list

    return [s for s in sprites_to_check if s.position == point]


def get_sprites_in_rect(rect: Rect, sprite_list: SpriteSequence[SpriteType]) -> list[SpriteType]:
    """
    Get a list of sprites in a particular rectangle. This function sees if any
    sprite overlaps the specified rectangle. If a sprite has a different
    center_x/center_y but touches the rectangle, this will return that sprite.

    The rectangle is specified as a tuple of (left, right, bottom, top).

    Args:
        rect: Rectangle to check
        sprite_list: SpriteList to check against

    Returns:
        List of sprites colliding, or an empty list.
    """
    if __debug__:
        if not isinstance(sprite_list, SpriteSequence):
            raise TypeError(f"Parameter 2 is a {type(sprite_list)} instead of expected SpriteList.")

    rect_points = rect.to_points()
    sprites_to_check: Iterable[SpriteType]

    if sprite_list.spatial_hash is not None:
        sprites_to_check = sprite_list.spatial_hash.get_sprites_near_rect(rect)
    else:
        sprites_to_check = sprite_list

    return [
        s
        for s in sprites_to_check
        if are_polygons_intersecting(rect_points, s.hit_box.get_adjusted_points())
    ]



from __future__ import annotations

from arcade.clock import GLOBAL_CLOCK, Clock


def boot_strap_clock(clock: Clock | None = None) -> Clock:
    """
    Because the sub_clock is not a fully featured part of Arcade we have to
    manipulate the clocks before the can be used with sub_clocks.

    This step will no longer be required when SubClocks become part of the main
    library.

    calling it will boostrap the global clock.
    DO NOT CALL MORE THAN ONCE PER CLOCK.

    Args:
        clock: a clcok that has yet to be setup. Defaults to arcade.clock.GLOBAL_CLOCK.
    """
    clock = clock or GLOBAL_CLOCK

    if hasattr(clock, "children"):
        raise ValueError(f"The clock {clock} has already been bootstrapped.")

    # No type check will ever like this, but we do what we must.
    clock.children = []  # type: ignore

    def recursive_tick(delta_time: float) -> None:
        clock.tick(delta_time)
        for child in clock.children:  # type: ignore
            child.tick(clock._tick_delta_time)

    # Hey we did a decorator manually! what a time to be alive.
    clock.tick = recursive_tick  # type: ignore

    def add_child(child: SubClock) -> None:
        clock.children.append(child)  # type: ignore

    clock.add_child = add_child  # type: ignore

    return clock


class SubClock(Clock):
    """
    A SubClock which gets ticked by a parent clock and can have its flow
    of time altered independantly of its parent or siblings.

    Args:
        parent: The clock which will tick the SubClock.
            could be the GLOBAL_CLOCK or another SubClock
        tick_speed: A multiplier on how the 'speed' of time.
            i.e. a value of 0.5 means time elapsed half as fast for this clock. Defaults to 1.0.
    """

    def __init__(self, parent: Clock | SubClock | None = None, tick_speed: float = 1) -> None:
        parent = parent or GLOBAL_CLOCK
        super().__init__(parent._elapsed_time, parent._tick, tick_speed)
        self.children: list[SubClock] = []
        try:
            parent.add_child(self)  # type: ignore
        except AttributeError:
            raise AttributeError(
                f"The clock {parent} has not been bootstrapped properly"
                f"call boot_strap_clock({parent}) before adding children"
            )

    def add_child(self, child: SubClock) -> None:
        self.children.append(child)

    def tick(self, delta_time: float) -> None:
        super().tick(delta_time)

        for child in self.children:
            child.tick(self.delta_time)
